/***************************************************************************
                        audio_block.h  -  description
                             -------------------
   Functions for handling AC3 audio blocks.
                             -------------------
    begin                : Fri Aug 29 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef _AUDIO_BLOCK_H_
#define _AUDIO_BLOCK_H_

#include "common/typedefs.h"
#include "AC3_audio.h"

#define AUDIO_BLOCK_read_config_STATE \
   (AC3_AUDIO_next_frame_STATE+(2*AC3_AUDIO_STATE_INC))
#define AUDIO_BLOCK_read_config_END_STATE \
   (AC3_AUDIO_next_frame_STATE+(3*AC3_AUDIO_STATE_INC))
#define AUDIO_BLOCK_read_exp_STATE \
   (AC3_AUDIO_next_frame_STATE+(3*AC3_AUDIO_STATE_INC))
#define AUDIO_BLOCK_read_exp_END_STATE \
   (AC3_AUDIO_next_frame_STATE+(4*AC3_AUDIO_STATE_INC))
#define AUDIO_BLOCK_read_bit_alloc_STATE \
   (AC3_AUDIO_next_frame_STATE+(4*AC3_AUDIO_STATE_INC))
#define AUDIO_BLOCK_read_bit_alloc_END_STATE \
   (AC3_AUDIO_next_frame_STATE+(5*AC3_AUDIO_STATE_INC))
#define AUDIO_BLOCK_read_mant_STATE \
   (AC3_AUDIO_next_frame_STATE+(5*AC3_AUDIO_STATE_INC))
#define AUDIO_BLOCK_read_mant_END_STATE \
   (AC3_AUDIO_next_frame_STATE+(6*AC3_AUDIO_STATE_INC))

#define AC3_MAX_CH           5
#define AC3_NUM_BINS       256
#define AC3_MAX_CPL_BIN    216
#define AC3_MAX_CH_BIN     253
#define AC3_MAX_LFE_BIN      7
#define AC3_NUM_BND         18
#define AC3_BIN_PER_SBND    12
#define AC3_BLKS_PER_FRAME   6
#define AC3_OUTPUT_BLK     0x80

#define AC3_DELTBAE_REUSE        0x00
#define AC3_DELTBAE_NEW          0x40
#define AC3_DELTBAE_NONE         0x80
#define AC3_DELTBAE_RESERVED     0xC0
#define AC3_DELTBAE(delt_array)  (*(delt_array) & 0xE0)
#define AC3_DELTNSEG(delt_array) (*(delt_array) & 0x1F)

#ifdef _NO_51_SURROUND_
# define AC3_OUTPUT_CHANNELS   2
#else
# define AC3_OUTPUT_CHANNELS   6
#endif
#define  AC3_COEFF_SIZE       16

#if (AC3_COEFF_SIZE == 32)
   typedef s32 AC3_COEFF;
   typedef s64 AC3_COEFF2;
# define CH_PAD_BYTES      1

#elif (AC3_COEFF_SIZE == 16)
   typedef s16 AC3_COEFF;
   typedef s32 AC3_COEFF2;
# define CH_PAD_BYTES      7

#endif

#define CPL_PAD_BYTES      2
#define LFE_PAD_BYTES      6
#define AC3_COEFF_SHIFT    (AC3_COEFF_SIZE-1)
#define AC3_COEFF_CONST(fpv) \
   (AC3_COEFF)(((fpv) * ((AC3_COEFF2)1 << (AC3_COEFF_SHIFT))) - 0.5)
#define AC3_COEFF_MULT_SHIFT(v1, v2, s) \
   (((AC3_COEFF2)(v1) * (AC3_COEFF2)(v2)) >> s)
#define AC3_COEFF_MULT(v1, v2) \
   (AC3_COEFF)AC3_COEFF_MULT_SHIFT(v1, v2, AC3_COEFF_SHIFT)

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
   u8 num_bins, mant_per_exp, snr_gain,
      delt[17],
      sdecay, fdecay, bai, csnroffst,
      bin_start, fs_leak, ch, pad1,
      exp[AC3_MAX_CPL_BIN+CPL_PAD_BYTES],
      bap[AC3_MAX_CPL_BIN+CPL_PAD_BYTES];
} CPL_STRUCT;

typedef struct {
   u8 num_bins, mant_per_exp, snr_gain,
      delt[17],
      exp[AC3_MAX_CH_BIN+CH_PAD_BYTES],
      bap[AC3_MAX_CH_BIN+CH_PAD_BYTES],
      cpl_coord[2+AC3_NUM_BND];
   AC3_COEFF coeff[AC3_NUM_BINS];
} CH_STRUCT;

typedef struct {
   u8 num_bins, mant_per_exp, snr_gain,
      pad1, pad2, pad3,
      exp[AC3_MAX_LFE_BIN+LFE_PAD_BYTES],
      bap[AC3_MAX_LFE_BIN+LFE_PAD_BYTES];
   AC3_COEFF coeff[AC3_MAX_LFE_BIN];
} LFE_STRUCT;

typedef struct {
   CPL_STRUCT  cpl_data;
   CH_STRUCT   ch_data[AC3_MAX_CH];
   LFE_STRUCT  lfe_data;
   LFE_STRUCT *lfe_ptr;
   u8         *u8_data;
   AC3_COEFF  *coeff;
   s16        *audio_buf[AC3_OUTPUT_CHANNELS],
               delay_buf[AC3_OUTPUT_CHANNELS*AC3_NUM_BINS];

   u32 num, rng, coupled_sbands, phase_invert;
   u8  /* dyn_range_gain[2], */
       blk_switch, dither, coupled_ch,
       num_coupled_bands, rematrix,
/*       gainrng[AC3_MAX_CH],*/
       prev_code[6],
       new_bai, ch, i, j;
} AUDIO_BLOCK;

/*
u8  CPL_STRUCT.num_bins,            ncplsubnd*12
    CPL_STRUCT.mant_per_exp,        cplexpstr
    CPL_STRUCT.snr_gain,      4     cplfsnroffst,
                              3     cplfgaincod
    CPL_STRUCT.delt[17],0     2     cpldeltbae,
                              3     cpldeltnseg
                        ODD   5     cpldeltoffst[seg]
                        EVEN  4     cpldeltlen[seg],
                              3     cpldeltba[seg]
    CPL_STRUCT.sdecay,              sdcycod
    CPL_STRUCT.fdecay,              fdcycod
    CPL_STRUCT.bai,           2     sgaincod,
                              2     dbpbcod,
                              3     floorcod
    CPL_STRUCT.csnroffst,     6     csnroffst
    CPL_STRUCT.bin_start,     4     cplbegf, cplstrtmant
    CPL_STRUCT.fs_leak,       3     cplfleak,
                              3     cplsleak
    CPL_STRUCT.ch,                  first_cpl_ch
    CPL_STRUCT.exp[216],      5/bin cpl_exps[bin]
    CPL_STRUCT.bap[216];            cpl_bap[bin]

u8  CH_STRUCT.num_bins,             endmant[ch]
    CH_STRUCT.mant_per_exp,         chexpstr[ch]
    CH_STRUCT.snr_gain,       4/ch  fsnroffst[ch],
                              3/ch  fgaincod[ch]
    CH_STRUCT.delt[17], 0     2/ch  deltbae[ch],
                              3/ch  deltnseg[ch]
                        ODD   5/seg deltoffst[ch][seg]
                        EVEN  4/seg deltlen[ch][seg],
                              3/seg deltba[ch][seg]
    CH_STRUCT.exp[253],       5/bin exps[ch][bin]
    CH_STRUCT.bap[253],             bap[bin]
    CH_STRUCT.cpl_coord[20];0       chincpl[ch],
                                    cplcoe[ch]
                            1 2     mstrcplco[ch]
                        2..19 4     cplcoexp[ch][bnd],
                              4     cplcomant[ch][bnd]

u8  LFE_STRUCT.num_bins,            7
    LFE_STRUCT.mant_per_exp,        lfeexpstr
    LFE_STRUCT.snr_gain,      4     lfefsnroffst,
                              3     lfefgaincod
    LFE_STRUCT.exp[7],        5/bin lfe_exps[bin]
    LFE_STRUCT.bap[7];              lfe_bap[bin]
   
u32 AUDIO_BLOCK.coupled_sbands,     1/bnd  cplbndstrc[bnd]
    AUDIO_BLOCK.phase_invert;       1/bnd  phsflg[bnd]
                                    1      phsflginu (set when phase_invert != 0)

u8  AUDIO_BLOCK.dyn_range_gain[],   8/ch   dynrng, dynrng2
    AUDIO_BLOCK.blk_switch,         1/ch   blksw[ch]
    AUDIO_BLOCK.dither,             1/ch   dithflag[ch]
    AUDIO_BLOCK.coupled_ch,         1/ch   chincpl[ch]
                                    1      cplinu (set when coupled_ch != 0)
    AUDIO_BLOCK.num_coupled_bands,         ncplbnd
    AUDIO_BLOCK.rematrix;           1/rbnd rematflg[rbnd]
                                    1      rematstr (set when rematrix != 0)
*/

/* Read the channel config info in the audio block..
      INPUT: frame - Pointer to an AC3_SYNC_FRAME. */
void AUDIO_BLOCK_read_config(AC3_SYNC_FRAME *frame,
                             AUDIO_BLOCK    *blk);

/* Read the exponents in the audio block..
      INPUT: frame - Pointer to an AC3_SYNC_FRAME. */
void AUDIO_BLOCK_read_exp(AC3_SYNC_FRAME *frame,
                          AUDIO_BLOCK    *blk);

/* Read the bit allocation info in the audio block..
      INPUT: frame - Pointer to an AC3_SYNC_FRAME. */
void AUDIO_BLOCK_read_bit_alloc(AC3_SYNC_FRAME *frame,
                                AUDIO_BLOCK    *blk);

/* Read the mantissas in the audio block..
      INPUT: frame - Pointer to an AC3_SYNC_FRAME. */
void AUDIO_BLOCK_read_mant(AC3_SYNC_FRAME *frame,
                           AUDIO_BLOCK    *blk);

#ifdef __cplusplus
};
#endif

#endif

