/***************************************************************************
                         mmx_imdft.c  -  description
                             -------------------
   MMX version of the AC-3 IMDFT process. Suitable for all processors 
   (PMMX, P2, and later) that support the MMX instruction set.
                             -------------------
    begin                : Tue Nov 29 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
   
   /* registers used: %r0   = data+? -> +4
                      %r1   = data+? -> -4
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code     src         dest */
# define STAGE1 \
      "movq "     "(%0), "    "%%mm0\n\t" \
      "movq "     "(%1), "    "%%mm1\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "movq "     "%%mm0, "   "%%mm3\n\t" \
      "pand "     "%%mm7, "   "%%mm2\n\t" \
      "pand "     "%%mm6, "   "%%mm3\n\t" \
      "movq "     "%%mm1, "   "%%mm4\n\t" \
      "movq "     "%%mm1, "   "%%mm5\n\t" \
      "pand "     "%%mm7, "   "%%mm4\n\t" \
      "pand "     "%%mm6, "   "%%mm5\n\t" \
      "paddw "    "%%mm2, "   "%%mm1\n\t" \
      "psubw "    "%%mm4, "   "%%mm0\n\t" \
      "psubw "    "%%mm3, "   "%%mm1\n\t" \
      "paddw "    "%%mm5, "   "%%mm0\n\t" \
      "movq "     "%%mm1, "   "(%0)\n\t" \
      "addl "     "$(4*2), "  "%0\n\t" \
      "movq "     "%%mm0, "   "(%1)\n\t" \
      "addl "     "$(4*2), "  "%1\n\t"

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code     src         dest */
# define STAGE_OUTPUT_1R2I(inc, r0, op0, r1, op1, off, d1, d2, sc) \
      "movq "     "(%" #r0 "), %%mm0\n\t" \
      "movq "     "%%mm6, "   "%%mm5\n\t" \
      "movq " #off"(%" #r0 "), %%mm1\n\t" \
      "ps" #d1 "lq $32, "     "%%mm5\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "movq "     "%%mm1, "   "%%mm3\n\t" \
      "movq "     "%%mm1, "   "%%mm4\n\t" \
      "pand "     "%%mm7, "   "%%mm2\n\t" \
      "pand "     "%%mm7, "   "%%mm3\n\t" \
      "pand "     "%%mm5, "   "%%mm4\n\t" \
      "ps" #d2 "lq $32, "     "%%mm5\n\t" \
      "por "      "%%mm2, "   "%%mm4\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "pand "     "%%mm5, "   "%%mm0\n\t" \
      "ps" #d2 "lq $" #sc   ", %%mm2\n\t" \
      "pand "     "%%mm5, "   "%%mm1\n\t" \
      "paddw "    "%%mm3, "   "%%mm0\n\t" \
      "psubw "    "%%mm3, "   "%%mm1\n\t" \
      "paddw "    "%%mm4, "   "%%mm0\n\t" \
      "paddw "    "%%mm4, "   "%%mm1\n\t" \
      "movq "     "%%mm7, "   "%%mm3\n\t" \
      "movq " #off"(%" #r1 "), %%mm4\n\t" \
      "psrlq "    "$16, "     "%%mm5\n\t" \
      "pand "     "%%mm4, "   "%%mm3\n\t" \
      "pand "     "%%mm7, "   "%%mm2\n\t" \
      "ps" #d1 "lq $" #sc   ", %%mm3\n\t" \
      "paddw "    "%%mm3, "   "%%mm0\n\t" \
      "psubw "    "%%mm3, "   "%%mm1\n\t" \
      "movq "     "%%mm0, "   "(%" #r0 ")\n\t" \
      "movq "     "(%" #r1 "), %%mm3\n\t" \
      "movq "     "%%mm1, "   #off "(%" #r0 ")\n\t" \
      "movq "     "%%mm4, "   "%%mm0\n\t" \
      "movq "     "%%mm3, "   "%%mm1\n\t" \
      "pand "     "%%mm6, "   "%%mm0\n\t" \
      "pand "     "%%mm5, "   "%%mm1\n\t" \
      "ps" #d1 "lq $32, "     "%%mm5\n\t" \
      "por "      "%%mm1, "   "%%mm0\n\t" \
      "movq "     "%%mm3, "   "%%mm1\n\t" \
      "pand "     "%%mm5, "   "%%mm3\n\t" \
      "pand "     "%%mm5, "   "%%mm4\n\t" \
      "pand "     "%%mm6, "   "%%mm1\n\t" \
      "paddw "    "%%mm0, "   "%%mm3\n\t" \
      "paddw "    "%%mm0, "   "%%mm4\n\t" \
      "psubw "    "%%mm1, "   "%%mm3\n\t" \
      "paddw "    "%%mm1, "   "%%mm4\n\t" \
      "paddw "    "%%mm2, "   "%%mm3\n\t" \
      "psubw "    "%%mm2, "   "%%mm4\n\t" \
      "movq "     "%%mm3, "   "(%" #r1 ")\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t" \
      "movq "     "%%mm4, "   #off "(%" #r1 ")\n\t" \
      #op1 "l $(" #inc "*2), ""%" #r1 "\n\t"

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code     src         dest */
# define STAGE_OUTPUT___2I(inc, r0, op0, r1, op1, off) \
      "movq "     "(%" #r0 "), %%mm0\n\t" \
      "movq " #off"(%" #r0 "), %%mm1\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "movq "     "%%mm1, "   "%%mm3\n\t" \
      "pand "     "%%mm7, "   "%%mm2\n\t" \
      "pand "     "%%mm7, "   "%%mm3\n\t" \
      "pand "     "%%mm6, "   "%%mm1\n\t" \
      "psubw "    "%%mm3, "   "%%mm2\n\t" \
      "paddw "    "%%mm3, "   "%%mm0\n\t" \
      "por "      "%%mm2, "   "%%mm1\n\t" \
      "movq "     "%%mm0, "   "(%" #r0 ")\n\t" \
      "movq "     "(%" #r1 "), %%mm5\n\t" \
      "movq "     "%%mm1, "   #off "(%" #r0 ")\n\t" \
      "movq " #off"(%" #r1 "), %%mm4\n\t" \
      "movq "     "%%mm5, "   "%%mm3\n\t" \
      "movq "     "%%mm4, "   "%%mm2\n\t" \
      "pand "     "%%mm6, "   "%%mm3\n\t" \
      "pand "     "%%mm6, "   "%%mm2\n\t" \
      "pand "     "%%mm7, "   "%%mm5\n\t" \
      "psubw "    "%%mm3, "   "%%mm2\n\t" \
      "paddw "    "%%mm3, "   "%%mm4\n\t" \
      "por "      "%%mm2, "   "%%mm5\n\t" \
      "movq "     "%%mm4, "   #off "(%" #r1 ")\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t" \
      "movq "     "%%mm5, "   "(%" #r1 ")\n\t" \
      #op1 "l $(" #inc "*2), ""%" #r1 "\n\t"

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code     src         dest */
# define STAGE_OUTPUT_2R2I(inc, r0, op0, r1, op1, off) \
      "movq "     "(%" #r0 "), %%mm0\n\t" \
      "movq " #off"(%" #r0 "), %%mm1\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "movq "     "%%mm0, "   "%%mm4\n\t" \
      "movq "     "%%mm1, "   "%%mm3\n\t" \
      "pand "     "%%mm6, "   "%%mm0\n\t" \
      "pand "     "%%mm7, "   "%%mm2\n\t" \
      "psllq "    "$16, "     "%%mm0\n\t" \
      "pand "     "%%mm6, "   "%%mm1\n\t" \
      "psrlq "    "$48, "     "%%mm4\n\t" \
      "por "      "%%mm2, "   "%%mm1\n\t" \
      "pand "     "%%mm7, "   "%%mm3\n\t" \
      "movq " #off"(%" #r1 "), %%mm5\n\t" \
      "movq "     "%%mm1, "   "%%mm2\n\t" \
      "paddw "    "%%mm3, "   "%%mm1\n\t" \
      "por "      "%%mm4, "   "%%mm0\n\t" \
      "movq "     "%%mm5, "   "%%mm4\n\t" \
      "pand "     "%%mm7, "   "%%mm5\n\t" \
      "psubw "    "%%mm3, "   "%%mm2\n\t" \
      "movq "     "%%mm5, "   "%%mm3\n\t" \
      "psllq "    "$48, "     "%%mm5\n\t" \
      "psrlq "    "$16, "     "%%mm3\n\t" \
      "pand "     "%%mm6, "   "%%mm4\n\t" \
      "por "      "%%mm5, "   "%%mm3\n\t" \
      "movq "     "(%" #r1 "), %%mm5\n\t" \
      "paddw "    "%%mm3, "   "%%mm1\n\t" \
      "psubw "    "%%mm3, "   "%%mm2\n\t" \
      "movq "     "%%mm5, "   "%%mm3\n\t" \
      "movq "     "%%mm1, "   "(%" #r0 ")\n\t" \
      "pand "     "%%mm7, "   "%%mm3\n\t" \
      "pand "     "%%mm6, "   "%%mm5\n\t" \
      "por "      "%%mm3, "   "%%mm4\n\t" \
      "movq "     "%%mm2, "   #off "(%" #r0 ")\n\t" \
      "movq "     "%%mm4, "   "%%mm3\n\t" \
      "psubw "    "%%mm5, "   "%%mm4\n\t" \
      "paddw "    "%%mm5, "   "%%mm3\n\t" \
      "paddw "    "%%mm0, "   "%%mm4\n\t" \
      "psubw "    "%%mm0, "   "%%mm3\n\t" \
      "movq "     "%%mm4, "   "(%" #r1 ")\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t" \
      "movq "     "%%mm3, "   #off "(%" #r1 ")\n\t" \
      #op1 "l $(" #inc "*2), ""%" #r1 "\n\t"

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code     src         dest */
# define STAGE_OUTPUT_2R__(inc, r0, op0, r1, op1, off) \
      "movq "     "(%" #r0 "), %%mm0\n\t" \
      "movq " #off"(%" #r0 "), %%mm1\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "pand "     "%%mm7, "   "%%mm0\n\t" \
      "movq "     "%%mm1, "   "%%mm3\n\t" \
      "pand "     "%%mm6, "   "%%mm2\n\t" \
      "pand "     "%%mm6, "   "%%mm3\n\t" \
      "movq "     "%%mm2, "   "%%mm5\n\t" \
      "psllq "    "$16, "     "%%mm2\n\t" \
      "movq " #off"(%" #r1 "), %%mm4\n\t" \
      "por "      "%%mm3, "   "%%mm0\n\t" \
      "movq "     "%%mm4, "   "%%mm3\n\t" \
      "psrlq "    "$48, "     "%%mm5\n\t" \
      "pand "     "%%mm7, "   "%%mm3\n\t" \
      "por "      "%%mm5, "   "%%mm2\n\t" \
      "movq "     "%%mm4, "   "%%mm5\n\t" \
      "psrlq "    "$16, "     "%%mm3\n\t" \
      "pand "     "%%mm6, "   "%%mm4\n\t" \
      "psllq "    "$48, "     "%%mm5\n\t" \
      "psubw "    "%%mm2, "   "%%mm4\n\t" \
      "por "      "%%mm5, "   "%%mm3\n\t" \
      "movq "     "(%" #r1 "), %%mm5\n\t" \
      "paddw "    "%%mm3, "   "%%mm0\n\t" \
      "psubw "    "%%mm3, "   "%%mm1\n\t" \
      "movq "     "%%mm0, "   "(%" #r0 ")\n\t" \
      "movq "     "%%mm5, "   "%%mm0\n\t" \
      "movq "     "%%mm1, "   #off "(%" #r0 ")\n\t" \
      "pand "     "%%mm7, "   "%%mm0\n\t" \
      "paddw "    "%%mm2, "   "%%mm5\n\t" \
      "paddw "    "%%mm0, "   "%%mm4\n\t" \
      "movq "     "%%mm5, "   "(%" #r1 ")\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t" \
      "movq "     "%%mm4, "   #off "(%" #r1 ")\n\t" \
      #op1 "l $(" #inc "*2), ""%" #r1 "\n\t"

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code     src         dest */
# define STAGE_OUTPUT_2R1I(inc, r0, op0, r1, op1, off, d1, d2) \
      "movq "     "%%mm7, "   "%%mm5\n\t" \
      "movq "     "(%" #r0 "), %%mm0\n\t" \
      "ps" #d2 "lq $32, "     "%%mm5\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "movq " #off"(%" #r0 "), %%mm3\n\t" \
      "pand "     "%%mm7, "   "%%mm0\n\t" \
      "movq "     "%%mm3, "   "%%mm1\n\t" \
      "pand "     "%%mm6, "   "%%mm3\n\t" \
      "movq "     "%%mm1, "   "%%mm4\n\t" \
      "por "      "%%mm3, "   "%%mm0\n\t" \
      "pand "     "%%mm5, "   "%%mm4\n\t" \
      "movq "     "%%mm0, "   "%%mm3\n\t" \
      "paddw "    "%%mm4, "   "%%mm0\n\t" \
      "pand "     "%%mm5, "   "%%mm3\n\t" \
      "ps" #d1 "lq $32, "     "%%mm5\n\t" \
      "psubw "    "%%mm4, "   "%%mm3\n\t" \
      "pcmpeqw "  "%%mm7, "   "%%mm5\n\t" \
      "pand "     "%%mm6, "   "%%mm2\n\t" \
      "pand "     "%%mm5, "   "%%mm1\n\t" \
      "por "      "%%mm3, "   "%%mm1\n\t" \
      "movq "     "%%mm2, "   "%%mm5\n\t" \
      "movq " #off"(%" #r1 "), %%mm4\n\t" \
      "psllq "    "$16, "     "%%mm2\n\t" \
      "movq "     "%%mm4, "   "%%mm3\n\t" \
      "psrlq "    "$48, "     "%%mm5\n\t" \
      "pand "     "%%mm7, "   "%%mm3\n\t" \
      "por "      "%%mm5, "   "%%mm2\n\t" \
      "movq "     "%%mm3, "   "%%mm5\n\t" \
      "psrlq "    "$16, "     "%%mm3\n\t" \
      "psllq "    "$48, "     "%%mm5\n\t" \
      "por "      "%%mm5, "   "%%mm3\n\t" \
      "movq "     "(%" #r1 "), %%mm5\n\t" \
      "paddw "    "%%mm3, "   "%%mm0\n\t" \
      "psubw "    "%%mm3, "   "%%mm1\n\t" \
      "movq "     "%%mm0, "   "(%" #r0 ")\n\t" \
      "movq "     "%%mm5, "   "%%mm3\n\t" \
      "movq "     "%%mm6, "   "%%mm0\n\t" \
      "pand "     "%%mm7, "   "%%mm3\n\t" \
      "movq "     "%%mm1, "   #off "(%" #r0 ")\n\t" \
      "paddw "    "%%mm2, "   "%%mm5\n\t" \
      "psubw "    "%%mm2, "   "%%mm3\n\t" \
      "movq "     "%%mm6, "   "%%mm2\n\t" \
      "pand "     "%%mm6, "   "%%mm4\n\t" \
      "ps" #d1 "lq $32, "     "%%mm0\n\t" \
      "por "      "%%mm3, "   "%%mm4\n\t" \
      "movq "     "%%mm5, "   "%%mm1\n\t" \
      "ps" #d2 "lq $32, "     "%%mm2\n\t" \
      "movq "     "%%mm4, "   "%%mm3\n\t" \
      "pcmpeqw "  "%%mm6, "   "%%mm2\n\t" \
      "pand "     "%%mm0, "   "%%mm1\n\t" \
      "pand "     "%%mm0, "   "%%mm3\n\t" \
      "paddw "    "%%mm1, "   "%%mm4\n\t" \
      "psubw "    "%%mm1, "   "%%mm3\n\t" \
      "pand "     "%%mm2, "   "%%mm5\n\t" \
      "movq "     "%%mm4, "   #off "(%" #r1 ")\n\t" \
      "por "      "%%mm3, "   "%%mm5\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t" \
      "movq "     "%%mm5, "   "(%" #r1 ")\n\t" \
      #op1 "l $(" #inc "*2), ""%" #r1 "\n\t"

   /* registers used: %r0   = data+? -> op0(inc)
                      %%mm4 = 0xFFFFFFFFFFFF0000 or 0x0000FFFFFFFFFFFF
                      %%mm5 = 0x000000000000FFFF or 0xFFFF000000000000 */
   /*  op_code     src         dest */
# define STAGE6_OUTPUT_01(inc, r0, op0, dat0, dat1) \
      "movq "     #dat0     ", %%mm0\n\t" \
      "movq "     #dat1     ", %%mm1\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "movq "     "%%mm1, "   "%%mm3\n\t" \
      "pand "     "%%mm4, "   "%%mm1\n\t" \
      "pand "     "%%mm5, "   "%%mm3\n\t" \
      "pand "     "%%mm5, "   "%%mm2\n\t" \
      "paddw "    "%%mm3, "   "%%mm0\n\t" \
      "psubw "    "%%mm3, "   "%%mm2\n\t" \
      "movq "     "%%mm0, "   #dat0 "\n\t" \
      "por "      "%%mm2, "   "%%mm1\n\t" \
      "movq "     "%%mm1, "   #dat1 "\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t"

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm4 = 0x000000000000FFFF
                      %%mm5 = 0xFFFF000000000000 */
   /*  op_code     src         dest */
# define STAGE6_OUTPUT_23(inc, r0, op0, r1, op1) \
      "movq "     "(%" #r1 "), %%mm0\n\t" \
      "movq "    "8(%" #r0 "), %%mm1\n\t" \
      "psrlq "    "$48, "     "%%mm0\n\t" \
      "movq "    "8(%" #r1 "), %%mm3\n\t" \
      "psllq "    "$48, "     "%%mm1\n\t" \
      "movq "     "(%" #r0 "), %%mm2\n\t" \
      "psubw "    "%%mm1, "   "%%mm3\n\t" \
      "paddw "    "%%mm0, "   "%%mm2\n\t" \
      "paddw "   "8(%" #r1 "), %%mm1\n\t" \
      "movq "     "%%mm3, "   "8(%" #r1 ")\n\t" \
      "pcmpgtw "  "%%mm3, "   "%%mm3\n\t" \
      "psubw "    "(%" #r0 "), %%mm0\n\t" \
      "pcmpeqw "  "%%mm5, "   "%%mm3\n\t" \
      "movq "     "%%mm2, "   "(%" #r0 ")\n\t" \
      "pcmpgtw "  "%%mm2, "   "%%mm2\n\t" \
      "pand "     "(%" #r1 "), %%mm3\n\t" \
      "pcmpeqw "  "%%mm4, "   "%%mm2\n\t" \
      "pand "     "%%mm5, "   "%%mm1\n\t" \
      "pand "     "%%mm4, "   "%%mm0\n\t" \
      "pand "    "8(%" #r0 "), %%mm2\n\t" \
      "por "      "%%mm1, "   "%%mm3\n\t" \
      "psubw "    "%%mm0, "   "%%mm2\n\t" \
      "movq "     "%%mm3, "   "(%" #r1 ")\n\t" \
      #op1 "l $(" #inc "*2), ""%" #r1 "\n\t" \
      "movq "     "%%mm2, "   "8(%" #r0 ")\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t"
   
   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc) */
   /*  op_code     src         dest */
# define STAGE7_OUTPUT(inc, r0, op0, r1, op1) \
      "movq "     "(%" #r0 "), %%mm2\n\t" \
      "pcmpgtd "  "%%mm3, "   "%%mm3\n\t" \
      "movq "     "(%" #r1 "), %%mm1\n\t" \
      "movq "     "%%mm2, "   "%%mm0\n\t" \
      "punpckhwd ""%%mm1, "   "%%mm2\n\t" \
      "psubw "    "%%mm0, "   "%%mm3\n\t" \
      "punpcklwd ""%%mm1, "   "%%mm0\n\t" \
      "movq "     "%%mm2, "   "%%mm4\n\t" \
      "punpcklwd ""%%mm1, "   "%%mm3\n\t" \
      "psllq "    "$32, "     "%%mm0\n\t" \
      "psllq "    "$32, "     "%%mm2\n\t" \
      "punpckhwd ""%%mm4, "   "%%mm0\n\t" \
      "psrlq "    "$32, "     "%%mm3\n\t" \
      "movq "     "%%mm0, "   "%%mm1\n\t" \
      "punpckldq ""%%mm0, "   "%%mm0\n\t" \
      "punpckhdq ""%%mm1, "   "%%mm1\n\t" \
      "psubw "    "%%mm2, "   "%%mm0\n\t" \
      "psrlq "    "$32, "     "%%mm2\n\t" \
      "psubw "    "%%mm3, "   "%%mm1\n\t" \
      "psllq "    "$32, "     "%%mm3\n\t" \
      "paddw "    "%%mm2, "   "%%mm0\n\t" \
      "paddw "    "%%mm3, "   "%%mm1\n\t" \
      "movq "     "%%mm0, "   "(%" #r0 ")\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t" \
      "movq "     "%%mm1, "   "(%" #r1 ")\n\t" \
      #op1 "l $(" #inc "*2), ""%" #r1 "\n\t"
   
   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code     src         dest */
# define SR_BF_k_eq_0_1I(inc, r0, op0, r1, op1, off, d1, d2) \
      "movq "     "%%mm7, "   "%%mm5\n\t" \
      "pxor "     "%%mm4, "   "%%mm4\n\t" \
      "ps" #d1 "lq $32, "     "%%mm5\n\t" \
      "movq "     "(%" #r0 "), %%mm0\n\t" \
      "pcmpeqw "  "%%mm5, "   "%%mm4\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "movq " #off"(%" #r0 "), %%mm3\n\t" \
      "pand "     "%%mm5, "   "%%mm2\n\t" \
      "movq "     "%%mm3, "   "%%mm1\n\t" \
      "pand "     "%%mm5, "   "%%mm3\n\t" \
      "pand "     "%%mm4, "   "%%mm1\n\t" \
      "movq "     "%%mm6, "   "%%mm5\n\t" \
      "psubw "    "%%mm3, "   "%%mm2\n\t" \
      "ps" #d2 "lq $32, "     "%%mm5\n\t" \
      "pxor "     "%%mm4, "   "%%mm4\n\t" \
      "paddw "    "%%mm3, "   "%%mm0\n\t" \
      "movq "     "(%" #r1 "), %%mm3\n\t" \
      "por "      "%%mm2, "   "%%mm1\n\t" \
      "movq "     "%%mm0, "   "(%" #r0 ")\n\t" \
      "movq " #off"(%" #r1 "), %%mm2\n\t" \
      "movq "     "%%mm3, "   "%%mm0\n\t" \
      "movq "     "%%mm1, "   #off "(%" #r0 ")\n\t" \
      "pcmpeqw "  "%%mm5, "   "%%mm4\n\t" \
      "movq "     "%%mm2, "   "%%mm1\n\t" \
      "pand "     "%%mm5, "   "%%mm3\n\t" \
      "pand "     "%%mm5, "   "%%mm2\n\t" \
      "pand "     "%%mm4, "   "%%mm0\n\t" \
      "paddw "    "%%mm3, "   "%%mm1\n\t" \
      "psubw "    "%%mm3, "   "%%mm2\n\t" \
      "movq "     "%%mm1, "   #off "(%" #r1 ")\n\t" \
      "por "      "%%mm2, "   "%%mm0\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t" \
      "movq "     "%%mm0, "   "(%" #r1 ")\n\t" \
      #op1 "l $(" #inc "*2), ""%" #r1 "\n\t"

   /* registers used: %r0    = data+? -> op0(inc)
                      %r1    = data+? -> op1(inc)
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code     src         dest */
# define SR_BF_k_eq_0_2I(inc, r0, op0, r1, op1, off) \
      "movq "     "(%" #r0 "), %%mm2\n\t" \
      "movq " #off"(%" #r0 "), %%mm3\n\t" \
      "movq "     "%%mm2, "   "%%mm0\n\t" \
      "movq "     "%%mm3, "   "%%mm1\n\t" \
      "pand "     "%%mm7, "   "%%mm2\n\t" \
      "pand "     "%%mm7, "   "%%mm3\n\t" \
      "movq "     "(%" #r1 "), %%mm4\n\t" \
      "pand "     "%%mm6, "   "%%mm1\n\t" \
      "psubw "    "%%mm3, "   "%%mm2\n\t" \
      "movq " #off"(%" #r1 "), %%mm5\n\t" \
      "paddw "    "%%mm3, "   "%%mm0\n\t" \
      "por "      "%%mm2, "   "%%mm1\n\t" \
      "movq "     "%%mm4, "   "%%mm3\n\t" \
      "movq "     "%%mm0, "   "(%" #r0 ")\n\t" \
      "movq "     "%%mm5, "   "%%mm2\n\t" \
      "pand "     "%%mm6, "   "%%mm3\n\t" \
      "movq "     "%%mm1, "   #off "(%" #r0 ")\n\t" \
      "pand "     "%%mm6, "   "%%mm2\n\t" \
      "pand "     "%%mm7, "   "%%mm4\n\t" \
      "psubw "    "%%mm3, "   "%%mm2\n\t" \
      "paddw "    "%%mm3, "   "%%mm5\n\t" \
      "por "      "%%mm2, "   "%%mm4\n\t" \
      "movq "     "%%mm5, "   #off "(%" #r1 ")\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t" \
      "movq "     "%%mm4, "   "(%" #r1 ")\n\t" \
      #op1 "l $(" #inc "*2), ""%" #r1 "\n\t"

   /* registers used: %4 = 0x5A815A81 */
   /*  op_code     src         dest */
# define SR_BF_k_eq_N8_SETUP \
      "movd "     "%4, "      "%%mm2\n\t" \
      "pxor "     "%%mm3, "   "%%mm3\n\t" \
      "punpcklwd ""%%mm2, "   "%%mm3\n\t"

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm3 = 0x5A8100005A810000
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code     src         dest */
# define SR_BF_k_eq_N8_1R(inc, r0, op0, r1, op1, off, d1, d2, sc1, sc2) \
      "movq " #off"(%" #r1 "), %%mm1\n\t" \
      "ps" #d2 "lq $32, "     "%%mm3\n\t" \
      "movq "     "(%" #r0 "), %%mm0\n\t" \
      "pand "     "%%mm7, "   "%%mm1\n\t" \
      "pand "     "%%mm6, "   "%%mm0\n\t" \
      "ps" #d2 "lq $" #sc1  ", %%mm1\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "movq " #off"(%" #r0 "), %%mm4\n\t" \
      "paddw "    "%%mm1, "   "%%mm0\n\t" \
      "movq "     "(%" #r1 "), %%mm5\n\t" \
      "psubw "    "%%mm2, "   "%%mm1\n\t" \
      "pand "     "%%mm7, "   "%%mm5\n\t" \
      "pand "     "%%mm6, "   "%%mm4\n\t" \
      "ps" #d2 "lq $" #sc1  ", %%mm5\n\t" \
      "movq "     "%%mm4, "   "%%mm2\n\t" \
      "psubw "    "%%mm5, "   "%%mm4\n\t" \
      "paddw "    "%%mm2, "   "%%mm5\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "paddw "    "%%mm4, "   "%%mm0\n\t" \
      "psubw "    "%%mm2, "   "%%mm4\n\t" \
      "pmulhw "   "%%mm3, "   "%%mm0\n\t" \
      "movq "     "%%mm1, "   "%%mm2\n\t" \
      "pmulhw "   "%%mm3, "   "%%mm4\n\t" \
      "paddw "    "%%mm5, "   "%%mm1\n\t" \
      "psllw "    "$1, "      "%%mm0\n\t" \
      "psubw "    "%%mm2, "   "%%mm5\n\t" \
      "psllw "    "$1, "      "%%mm4\n\t" \
      "movq "     "%%mm3, "   "%%mm2\n\t" \
      "pmulhw "   "%%mm3, "   "%%mm1\n\t" \
      "psrlw "    "$14, "     "%%mm2\n\t" \
      "pmulhw "   "%%mm3, "   "%%mm5\n\t" \
      "movq "     "%%mm2, "   "%%mm3\n\t" \
      "ps" #d1 "lq $" #sc1  ", %%mm2\n\t" \
      "por "      "%%mm3, "   "%%mm0\n\t" \
      "ps" #d1 "lq $" #sc2  ", %%mm1\n\t" \
      "por "      "%%mm3, "   "%%mm4\n\t" \
      "ps" #d1 "lq $" #sc2  ", %%mm5\n\t" \
      "por "      "%%mm2, "   "%%mm1\n\t" \
      "por "      "%%mm2, "   "%%mm5\n\t" \
      "pxor "     "%%mm2, "   "%%mm2\n\t" \
      "pand "     "%%mm7, "   "%%mm1\n\t" \
      "pcmpeqw "  "%%mm3, "   "%%mm2\n\t" \
      "movq "     "(%" #r0 "), %%mm3\n\t" \
      "pand "     "%%mm7, "   "%%mm5\n\t" \
      "pand "     "%%mm2, "   "%%mm3\n\t" \
      "pand " #off"(%" #r0 "), %%mm2\n\t" \
      "por "      "%%mm3, "   "%%mm0\n\t" \
      "movq "     "%%mm7, "   "%%mm3\n\t" \
      "por "      "%%mm2, "   "%%mm4\n\t" \
      "ps" #d1 "lq $32, "     "%%mm3\n\t" \
      "pxor "     "%%mm2, "   "%%mm2\n\t" \
      "movq "     "%%mm0, "   "(%" #r0 ")\n\t" \
      "pcmpeqw "  "%%mm2, "   "%%mm3\n\t" \
      "movq "     "(%" #r1 "), %%mm0\n\t" \
      "movq "     "%%mm4, "   #off "(%" #r0 ")\n\t" \
      "pand "     "%%mm3, "   "%%mm0\n\t" \
      "pand " #off"(%" #r1 "), %%mm3\n\t" \
      "por "      "%%mm0, "   "%%mm5\n\t" \
      "por "      "%%mm3, "   "%%mm1\n\t" \
      "movq "     "%%mm5, "   "(%" #r1 ")\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t" \
      "movq "     "%%mm1, "   #off "(%" #r1 ")\n\t" \
      #op1 "l $(" #inc "*2), ""%" #r1 "\n\t"

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %%mm3 = 0x5A8100005A810000
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code     src         dest */
# define SR_BF_k_eq_N8_2R(inc, r0, op0, r1, op1, off) \
      "movq " #off"(%" #r1 "), %%mm1\n\t" \
      "movq "     "(%" #r0 "), %%mm0\n\t" \
      "movq "     "%%mm1, "   "%%mm2\n\t" \
      "pand "     "%%mm7, "   "%%mm1\n\t" \
      "psllq "    "$48, "     "%%mm2\n\t" \
      "movq "     "(%" #r1 "), %%mm5\n\t" \
      "psrlq "    "$16, "     "%%mm1\n\t" \
      "pand "     "%%mm6, "   "%%mm0\n\t" \
      "por "      "%%mm2, "   "%%mm1\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "paddw "    "%%mm1, "   "%%mm0\n\t" \
      "psubw "    "%%mm2, "   "%%mm1\n\t" \
      "movq "     "%%mm5, "   "%%mm2\n\t" \
      "pand "     "%%mm7, "   "%%mm5\n\t" \
      "psllq "    "$48, "     "%%mm2\n\t" \
      "movq " #off"(%" #r0 "), %%mm4\n\t" \
      "psrlq "    "$16, "     "%%mm5\n\t" \
      "pand "     "%%mm6, "   "%%mm4\n\t" \
      "por "      "%%mm2, "   "%%mm5\n\t" \
      "movq "     "%%mm4, "   "%%mm2\n\t" \
      "psubw "    "%%mm5, "   "%%mm4\n\t" \
      "paddw "    "%%mm2, "   "%%mm5\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "paddw "    "%%mm4, "   "%%mm0\n\t" \
      "psubw "    "%%mm2, "   "%%mm4\n\t" \
      "pmulhw "   "%%mm3, "   "%%mm0\n\t" \
      "movq "     "%%mm1, "   "%%mm2\n\t" \
      "pmulhw "   "%%mm3, "   "%%mm4\n\t" \
      "paddw "    "%%mm5, "   "%%mm1\n\t" \
      "psubw "    "%%mm2, "   "%%mm5\n\t" \
      "pmulhw "   "%%mm3, "   "%%mm1\n\t" \
      "movq "     "%%mm3, "   "%%mm2\n\t" \
      "psllw "    "$1, "      "%%mm0\n\t" \
      "psrlw "    "$14, "     "%%mm2\n\t" \
      "psllw "    "$1, "      "%%mm4\n\t" \
      "por "      "%%mm2, "   "%%mm0\n\t" \
      "pmulhw "   "%%mm3, "   "%%mm5\n\t" \
      "por "      "%%mm2, "   "%%mm4\n\t" \
      "psllw "    "$1, "      "%%mm1\n\t" \
      "psllw "    "$1, "      "%%mm5\n\t" \
      "por "      "%%mm2, "   "%%mm1\n\t" \
      "por "      "%%mm2, "   "%%mm5\n\t" \
      "movq "     "(%" #r0 "), %%mm2\n\t" \
      "pand "     "%%mm7, "   "%%mm2\n\t" \
      "por "      "%%mm2, "   "%%mm0\n\t" \
      "movq " #off"(%" #r0 "), %%mm2\n\t" \
      "movq "     "%%mm0, "   "(%" #r0 ")\n\t" \
      "pand "     "%%mm7, "   "%%mm2\n\t" \
      "movq "     "%%mm5, "   "%%mm0\n\t" \
      "psllq "    "$16, "     "%%mm5\n\t" \
      "por "      "%%mm2, "   "%%mm4\n\t" \
      "psrlq "    "$48, "     "%%mm0\n\t" \
      "movq "     "%%mm4, "   #off "(%" #r0 ")\n\t" \
      "por "      "%%mm0, "   "%%mm5\n\t" \
      "movq "     "%%mm1, "   "%%mm4\n\t" \
      "psllq "    "$16, "     "%%mm1\n\t" \
      "movq "     "(%" #r1 "), %%mm0\n\t" \
      "psrlq "    "$48, "     "%%mm4\n\t" \
      "pand "     "%%mm6, "   "%%mm0\n\t" \
      "movq " #off"(%" #r1 "), %%mm2\n\t" \
      "por "      "%%mm4, "   "%%mm1\n\t" \
      "pand "     "%%mm6, "   "%%mm2\n\t" \
      "por "      "%%mm0, "   "%%mm5\n\t" \
      "por "      "%%mm2, "   "%%mm1\n\t" \
      "movq "     "%%mm5, "   "(%" #r1 ")\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t" \
      "movq "     "%%mm1, "   #off "(%" #r1 ")\n\t" \
      #op1 "l $(" #inc "*2), ""%" #r1 "\n\t"

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %2    = IDFT_CONST
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code     src         dest */
# define SR_BF_1R(inc, r0, op0, r1, op1, off, h1, h2, d1, d2) \
      "movq "     "(%" #r0 "), %%mm0\n\t" \
      "movq " #off"(%" #r1 "), %%mm2\n\t" \
      "pand "     "%%mm6, "   "%%mm0\n\t" \
      "movq " #off"(%" #r0 "), %%mm1\n\t" \
      "pand "     "%%mm7, "   "%%mm2\n\t" \
      "movq "     "(%" #r1 "), %%mm3\n\t" \
  "punpck" #h1 "dq %%mm0, "   "%%mm0\n\t" \
      "pand "     "%%mm6, "   "%%mm1\n\t" \
  "punpck" #h2 "dq %%mm2, "   "%%mm2\n\t" \
      "pand "     "%%mm7, "   "%%mm3\n\t" \
  "punpck" #h1 "dq %%mm1, "   "%%mm1\n\t" \
      "movq "     "(%2), "    "%%mm4\n\t" \
  "punpck" #h2 "dq %%mm3, "   "%%mm3\n\t" \
      "por "      "%%mm0, "   "%%mm2\n\t" \
      "movq "     "8(%2), "   "%%mm5\n\t" \
      "pmaddwd "  "%%mm4, "   "%%mm2\n\t" \
      "por "      "%%mm3, "   "%%mm1\n\t" \
      "movq "     "%%mm6, "   "%%mm4\n\t" \
      "pmaddwd "  "%%mm5, "   "%%mm1\n\t" \
      "movq "     "%%mm7, "   "%%mm5\n\t" \
      "movq "     "%%mm1, "   "%%mm0\n\t" \
      "psrld "    "$2, "      "%%mm5\n\t" \
      "paddd "    "%%mm2, "   "%%mm1\n\t" \
      "psubd "    "%%mm2, "   "%%mm0\n\t" \
      "paddd "    "%%mm5, "   "%%mm1\n\t" \
      "paddd "    "%%mm5, "   "%%mm0\n\t" \
      "pslld "    "$1, "      "%%mm1\n\t" \
      "pxor "     "%%mm3, "   "%%mm3\n\t" \
      "pslld "    "$1, "      "%%mm0\n\t" \
      "movq "     "(%" #r0 "), %%mm2\n\t" \
      "ps" #d1 "lq $32, "     "%%mm4\n\t" \
      "movq "     "%%mm1, "   "%%mm5\n\t" \
      "psrlq "    "$16, "     "%%mm1\n\t" \
      "pcmpeqw "  "%%mm4, "   "%%mm3\n\t" \
      "pand "     "%%mm4, "   "%%mm5\n\t" \
      "pand "     "%%mm3, "   "%%mm2\n\t" \
      "pand "     "%%mm0, "   "%%mm4\n\t" \
      "pand " #off"(%" #r0 "), %%mm3\n\t" \
      "psrlq "    "$16, "     "%%mm0\n\t" \
      "por "      "%%mm4, "   "%%mm2\n\t" \
      "por "      "%%mm3, "   "%%mm5\n\t" \
      "movq "     "%%mm7, "   "%%mm4\n\t" \
      "movq "     "%%mm2, "   "(%" #r0 ")\n\t" \
      "ps" #d2 "lq $32, "     "%%mm4\n\t" \
      "pxor "     "%%mm3, "   "%%mm3\n\t" \
      "movq "     "%%mm5, "   #off "(%" #r0 ")\n\t" \
      "pcmpeqw "  "%%mm4, "   "%%mm3\n\t" \
      "movq " #off"(%" #r1 "), %%mm2\n\t" \
      "pand "     "%%mm4, "   "%%mm0\n\t" \
      "pand "     "%%mm4, "   "%%mm1\n\t" \
      "pand "     "%%mm3, "   "%%mm2\n\t" \
      "pand "     "(%" #r1 "), %%mm3\n\t" \
      "por "      "%%mm2, "   "%%mm0\n\t" \
      "por "      "%%mm3, "   "%%mm1\n\t" \
      "movq "     "%%mm0, "   #off "(%" #r1 ")\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t" \
      "movq "     "%%mm1, "   "(%" #r1 ")\n\t" \
      #op1 "l $(" #inc "*2), ""%" #r1 "\n\t"

   /* registers used: %r0   = data+? -> op0(inc)
                      %r1   = data+? -> op1(inc)
                      %2    = IDFT_CONST
                      %%mm6 = 0xFFFF0000FFFF0000
                      %%mm7 = 0x0000FFFF0000FFFF */
   /*  op_code     src         dest */
# define SR_BF_2R(inc, r0, op0, r1, op1, off, h1, h2) \
      "movq "     "(%" #r0 "), %%mm0\n\t" \
      "movq " #off"(%" #r1 "), %%mm1\n\t" \
      "pand "     "%%mm6, "   "%%mm0\n\t" \
      "pand "     "%%mm7, "   "%%mm1\n\t" \
      "movq "     "(%" #r1 "), %%mm3\n\t" \
      "movq "     "%%mm1, "   "%%mm2\n\t" \
      "pand "     "%%mm7, "   "%%mm3\n\t" \
      "psllq "    "$32, "     "%%mm1\n\t" \
      "movq "     "%%mm3, "   "%%mm5\n\t" \
      "psrlq "    "$32, "     "%%mm2\n\t" \
      "por "      "%%mm1, "   "%%mm0\n\t" \
      "psllq "    "$32, "     "%%mm3\n\t" \
      "por "      "%%mm2, "   "%%mm0\n\t" \
      "movq "     "(%2), "    "%%mm4\n\t" \
      "psrlq "    "$32, "     "%%mm5\n\t" \
      "movq " #off"(%" #r0 "), %%mm2\n\t" \
      "por "      "%%mm5, "   "%%mm3\n\t" \
      "movq "     "%%mm4, "   "%%mm5\n\t" \
"punpck" #h1 "dq ""%%mm4, "   "%%mm4\n\t" \
      "pand "     "%%mm6, "   "%%mm2\n\t" \
"punpck" #h2 "dq ""%%mm5, "   "%%mm5\n\t" \
      "movq "     "%%mm0, "   "%%mm1\n\t" \
      "pmaddwd "  "%%mm4, "   "%%mm0\n\t" \
      "movq "     "8(%2), "   "%%mm4\n\t" \
      "por "      "%%mm3, "   "%%mm2\n\t" \
      "pmaddwd "  "%%mm5, "   "%%mm1\n\t" \
      "movq "     "%%mm4, "   "%%mm5\n\t" \
"punpck" #h1 "dq ""%%mm4, "   "%%mm4\n\t" \
      "movq "     "%%mm2, "   "%%mm3\n\t" \
      "pmaddwd "  "%%mm4, "   "%%mm2\n\t" \
"punpck" #h2 "dq ""%%mm5, "   "%%mm5\n\t" \
      "movq "     "%%mm7, "   "%%mm4\n\t" \
      "pmaddwd "  "%%mm5, "   "%%mm3\n\t" \
      "psrlw "    "$2, "      "%%mm4\n\t" \
      "paddd "    "%%mm4, "   "%%mm3\n\t" \
      "paddd "    "%%mm4, "   "%%mm2\n\t" \
      "movq "     "%%mm3, "   "%%mm5\n\t" \
      "psubd "    "%%mm1, "   "%%mm3\n\t" \
      "paddd "    "%%mm5, "   "%%mm1\n\t" \
      "movq "     "%%mm2, "   "%%mm4\n\t" \
      "movq "     "%%mm1, "   "%%mm5\n\t" \
      "psubd "    "%%mm0, "   "%%mm2\n\t" \
      "psllq "    "$17, "     "%%mm1\n\t" \
      "paddd "    "%%mm4, "   "%%mm0\n\t" \
      "psrlq "    "$47, "     "%%mm5\n\t" \
      "movq "     "%%mm3, "   "%%mm4\n\t" \
      "pslld "    "$1, "      "%%mm2\n\t" \
      "por "      "%%mm5, "   "%%mm1\n\t" \
      "psllq "    "$17, "     "%%mm3\n\t" \
      "movq " #off"(%" #r0 "), %%mm5\n\t" \
      "pslld "    "$1, "      "%%mm0\n\t" \
      "pand "     "%%mm6, "   "%%mm2\n\t" \
      "psrlq "    "$47, "     "%%mm4\n\t" \
      "pand "     "%%mm6, "   "%%mm0\n\t" \
      "pand "     "%%mm7, "   "%%mm5\n\t" \
      "por "      "%%mm4, "   "%%mm3\n\t" \
      "movq "     "(%" #r0 "), %%mm4\n\t" \
      "por "      "%%mm5, "   "%%mm0\n\t" \
      "pand "     "%%mm7, "   "%%mm4\n\t" \
      "movq "     "%%mm0, "   #off "(%" #r0 ")\n\t" \
      "por "      "%%mm4, "   "%%mm2\n\t" \
      "movq "     "(%" #r1 "), %%mm0\n\t" \
      "pand "     "%%mm7, "   "%%mm1\n\t" \
      "pand "     "%%mm6, "   "%%mm0\n\t" \
      "movq "     "%%mm2, "   "(%" #r0 ")\n\t" \
      "movq " #off"(%" #r1 "), %%mm2\n\t" \
      "pand "     "%%mm7, "   "%%mm3\n\t" \
      "pand "     "%%mm6, "   "%%mm2\n\t" \
      "por "      "%%mm0, "   "%%mm1\n\t" \
      "por "      "%%mm2, "   "%%mm3\n\t" \
      "movq "     "%%mm1, "   "(%" #r1 ")\n\t" \
      #op0 "l $(" #inc "*2), ""%" #r0 "\n\t" \
      "movq "     "%%mm3, "   #off "(%" #r1 ")\n\t" \
      #op1 "l $(" #inc "*2), ""%" #r1 "\n\t"

   /* registers used: %r0 = data+? -> add(%4)
                      %2  = IDFT_CONST
                      %4  = -> (%2) */
   /*  op_code     src         dest */
# define SWAP8_1(r0) \
      "movq "     "(%" #r0 "), %%mm0\n\t" \
      "movq "   "32(%" #r0 "), %%mm1\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "movq "  "256(%" #r0 "), %%mm4\n\t" \
      "movq "     "%%mm1, "   "%%mm3\n\t" \
      "punpckldq ""%%mm4, "   "%%mm0\n\t" \
      "movq "  "288(%" #r0 "), %%mm5\n\t" \
      "movq "     "%%mm0, "   "(%" #r0 ")\n\t" \
      "punpckldq ""%%mm5, "   "%%mm1\n\t" \
      "movl "     "(%2), "    "%4\n\t" \
      "punpckhdq ""%%mm4, "   "%%mm2\n\t" \
      "movq "     "%%mm1, "   "32(%" #r0 ")\n\t" \
      "punpckhdq ""%%mm5, "   "%%mm3\n\t" \
      "movq "     "%%mm2, "   "256(%" #r0 ")\n\t" \
      "addl "     "$4, "      "%2\n\t" \
      "movq "     "%%mm3, "   "288(%" #r0 ")\n\t" \
      "addl "     "%4, "      "%" #r0 "\n\t"

   /* registers used: %r0 = data+? -> sub(%4)
                      %r1 = data+? -> sub(%4)
                      %2  = IDFT_CONST
                      %4  = -> (%2) */
   /*  op_code     src         dest */
# define SWAP8_2(r0, r1) \
      "movq "     "(%" #r1 "), %%mm0\n\t" \
      "movq "     "(%" #r0 "), %%mm1\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "movq "  "256(%" #r1 "), %%mm4\n\t" \
      "movq "     "%%mm1, "   "%%mm3\n\t" \
      "punpckldq ""%%mm4, "   "%%mm0\n\t" \
      "movq "  "256(%" #r0 "), %%mm5\n\t" \
      "movq "     "%%mm0, "   "(%" #r0 ")\n\t" \
      "punpckldq ""%%mm5, "   "%%mm1\n\t" \
      "movl "     "(%2), "    "%4\n\t" \
      "punpckhdq ""%%mm4, "   "%%mm2\n\t" \
      "movq "     "%%mm1, "   "(%" #r1 ")\n\t" \
      "punpckhdq ""%%mm5, "   "%%mm3\n\t" \
      "movq "     "%%mm2, "   "256(%" #r0 ")\n\t" \
      "subl "     "%4, "      "%" #r0 "\n\t" \
      "movl "     "4(%2), "    "%4\n\t" \
      "addl "     "$8, "      "%2\n\t" \
      "movq "     "%%mm3, "   "256(%" #r1 ")\n\t" \
      "subl "     "%4, "      "%" #r1 "\n\t"

   /* registers used: %0   = data+? -> add(inc)
                      %1   = data+? -> sub(inc)
                      %2   = IDFT_CONST
                      %mm7 = 0x0001000100010001 */
   /*  op_code     src         dest */
# define PRE_POST_SCALING(inc, m1, m2) \
      "movq "     "(%0), "    "%%mm0\n\t" \
      "movq "     "(%1), "    "%%mm1\n\t" \
      "movq "     "%%mm0, "   "%%mm4\n\t" \
      "movq "     "%%mm0, "   "%%mm2\n\t" \
      "psllq "    "$32, "     "%%mm4\n\t" \
      "movq "     "%%mm1, "   "%%mm5\n\t" \
      "punpckhwd ""%%mm4, "   "%%mm2\n\t" \
      "movq "     "%%mm1, "   "%%mm3\n\t" \
      "psllq "    "$32, "     "%%mm5\n\t" \
      "movq "     "%%mm2, "   "%%mm4\n\t" \
      "punpckhwd ""%%mm5, "   "%%mm3\n\t" \
      "psllq "    "$32, "     "%%mm4\n\t" \
      "movq "     "%%mm3, "   "%%mm5\n\t" \
      "punpckhwd ""%%mm4, "   "%%mm2\n\t" \
      "psllq "    "$32, "     "%%mm5\n\t" \
      "movq "     "(%2), "    "%%mm4\n\t" \
      "punpckhwd ""%%mm5, "   "%%mm3\n\t" \
      "movq "     "8(%2), "   "%%mm5\n\t" \
      "pmulhw "   "%%mm4, "   "%%mm0\n\t" \
      "movq "     "%%mm4, "   "%%mm6\n\t" \
      "pmulhw "   "%%mm5, "   "%%mm3\n\t" \
      "psllq "    "$32, "     "%%mm6\n\t" \
      "paddw "    "%%mm3, "   "%%mm0\n\t" \
      "punpckhwd ""%%mm6, "   "%%mm4\n\t" \
      "movq "     "%%mm5, "   "%%mm3\n\t" \
      "movq "     "%%mm4, "   "%%mm6\n\t" \
      "psllq "    "$32, "     "%%mm3\n\t" \
      "psllq "    "$32, "     "%%mm6\n\t" \
      "punpckhwd ""%%mm3, "   "%%mm5\n\t" \
      "punpckhwd ""%%mm6, "   "%%mm4\n\t" \
      "movq "     "%%mm5, "   "%%mm3\n\t" \
      "pmulhw "   "%%mm4, "   "%%mm1\n\t" \
      "psllq "    "$32, "     "%%mm3\n\t" \
      "punpckhwd ""%%mm3, "   "%%mm5\n\t" \
      "paddw "    "%%mm7, "   "%%mm0\n\t" \
      "pmulhw "   "%%mm5, "   "%%mm2\n\t" \
      "addl "     "$16, "     "%2\n\t" \
      "psllw "    "$1, "      "%%mm0\n\t" \
      "psubw "    "%%mm"#m2 ", %%mm" #m1 "\n\t" \
      "movq "     "%%mm0, "   "(%0)\n\t" \
      "psllw "    "$1, "      "%%mm" #m1 "\n\t" \
      "addl $("   #inc "*2), ""%0\n\t" \
      "movq "     "%%mm"#m1 ", (%1)\n\t" \
      "subl $("   #inc "*2), ""%1\n\t"


void AC3_imdft512(AC3_COEFF *data)
{
   __asm__ __volatile__(
   /* Pre-IDFT Scaling:
      registers used: %0 = data
                      %1 = data+252
                      %3 = 32 */
   /*  op_code     src         dest */
      "pcmpeqd "  "%%mm7, "   "%%mm7\n\t"
      "psrlw "    "$15, "     "%%mm7\n\t"    /* mm7 = 0x0001000100010001 */
      /* %0 = data      %1 = data+252 */
      "PRE_LOOP:\n\t"
      PRE_POST_SCALING(4, 1, 2)
      "dec "                  "%3\n\t"
      "jnz "      "PRE_LOOP\n\t"
      /* %0 = data+128  %1 = data+124 */

   /* INIT MASKS */
   /*  op_code     src         dest */
      "pcmpeqd "  "%%mm7, "   "%%mm7\n\t"
      "pxor "     "%%mm6, "   "%%mm6\n\t"
      "movq "     "%%mm7, "   "%%mm5\n\t"
      "punpcklwd ""%%mm6, "   "%%mm7\n\t"    /* mm7 = 0x0000FFFF0000FFFF */
      "punpcklwd ""%%mm5, "   "%%mm6\n\t"    /* mm6 = 0xFFFF0000FFFF0000 */
   
   /* STAGE 1: 2pt-DFTs
      registers used: %0 = data
                      %1 = data+128 
                      %3 = 32 */
   /*  op_code     src         dest */
      "subl "     "$(128*2), ""%0\n\t"
      "addl "     "$(4*2), "  "%1\n\t"
      "addl "     "$32, "     "%3\n\t"
      /* %0 = data      %1 = data+128 */
      "STAGE1_LOOP:\n\t"
      STAGE1
      "dec "                  "%3\n\t"
      "jnz "      "STAGE1_LOOP\n\t"
      /* %0 = data+128  %1 = data+256 */

   /* STAGE 2: 4pt-DFTs
      registers used: %0 = data
                      %1 = data+188
                      %3 = 4 */
   /*  op_code     src         dest */
      "subl "     "$(128*2), ""%0\n\t"
      "subl "     "$(68*2), " "%1\n\t"
      "addl "     "$4, "      "%3\n\t"
      /* %0 = data      %1 = data+188 */
      STAGE_OUTPUT_1R2I(4, 0, add, 1, sub, 128, l, r, 48)
      STAGE_OUTPUT___2I(4, 0, add, 1, sub, 128)
      STAGE_OUTPUT_2R2I(4, 0, add, 1, sub, 128)
      STAGE_OUTPUT_2R2I(4, 0, add, 1, sub, 128)
      "STAGE2_LOOP1:\n\t"
      STAGE_OUTPUT___2I(4, 0, add, 1, sub, 128)
      "dec "                  "%3\n\t"
      "jnz "      "STAGE2_LOOP1\n\t"
      "addl "     "$4, "      "%3\n\t"
      "STAGE2_LOOP2:\n\t"
      STAGE_OUTPUT_2R__(4, 0, add, 1, sub, 128)
      "dec "                  "%3\n\t"
      "jnz "      "STAGE2_LOOP2\n\t"
      STAGE_OUTPUT_2R2I(4, 0, add, 1, sub, 128)
      STAGE_OUTPUT_2R2I(4, 0, add, 1, sub, 128)
      STAGE_OUTPUT_2R__(4, 0, add, 1, sub, 128)
      STAGE_OUTPUT_2R1I(0, 0, add, 1, sub, 128, l, r)
      /* %0 = data+60   %1 = data+128 */

   /* STAGE 3: 8pt-DFTs
      registers used: %0 = data+60
                      %1 = data+128
                      %3 = 4
                      %4 = 0x5A815A81 */
   /*  op_code     src         dest */
      "addl "     "$4, "      "%3\n\t"
      /* 1, 3 */
      /* %0 = data+60   %1 = data+128 */
      SR_BF_k_eq_0_1I( 4, 0, sub, 1, add, 128, l, r)
      SR_BF_k_eq_0_2I(12, 0, sub, 1, add, 128)
      "STAGE3_LOOP1:\n\t"
      SR_BF_k_eq_0_2I( 4, 0, sub, 1, add, 128)
      "dec "                  "%3\n\t"
      "jnz "      "STAGE3_LOOP1\n\t"
      /* 7, 5 */
      /* %0 = data+28   %1 = data+160 */
      "addl "     "$4, "      "%3\n\t"
      SR_BF_k_eq_N8_SETUP
      "STAGE3_LOOP2:\n\t"
      SR_BF_k_eq_N8_2R(4, 0, sub, 1, add, 128)
      "dec "                  "%3\n\t"
      "jnz "      "STAGE3_LOOP2\n\t"
      "subl "     "$(8*2), "  "%0\n\t"
      "addl "     "$(8*2), "  "%1\n\t"
      SR_BF_k_eq_N8_2R(4, 0, sub, 1, add, 128)
      SR_BF_k_eq_N8_1R(0, 0, sub, 1, add, 128, l, r, 16, 17)
      /* %0 = data      %1 = data+188 */
      
      "addl "     "$(32*2), " "%1\n\t"
      "addl "     "$2, "      "%3\n\t"
      /* %0 = data      %1 = data+220 */
      "STAGE3_LOOP3:\n\t"
      STAGE_OUTPUT_1R2I(  4, 0, add, 1, sub, 64, r, l, 16)
      STAGE_OUTPUT_2R2I(  4, 0, add, 1, sub, 64)
      STAGE_OUTPUT___2I(  4, 0, add, 1, sub, 64)
      STAGE_OUTPUT___2I(  4, 0, add, 1, sub, 64)
      STAGE_OUTPUT_2R__(  4, 0, add, 1, sub, 64)
      STAGE_OUTPUT_2R__(  4, 0, add, 1, sub, 64)
      STAGE_OUTPUT_2R2I(  4, 0, add, 1, sub, 64)
      STAGE_OUTPUT_2R1I(100, 0, add, 1, sub, 64, r, l)
      "dec "                  "%3\n\t"
      "jnz "      "STAGE3_LOOP3\n\t"
      /* %0 = data+256  %1 = data-36 */

   /* STAGE 4: 16pt-DFTs
      registers used: %0 = data
                      %1 = data+220
                      %2 = IDFT_CONST
                      %3 = 4
                      %4 = 0x5A815A81 */
   /*  op_code     src         dest */
      "subl "     "$(256*2), ""%0\n\t"
      "addl "     "$(256*2), ""%1\n\t"
      "addl "     "$4, "      "%3\n\t"
      /* 15, 13 */
      /* %0 = data      %1 = data+220 */
      SR_BF_1R( 8, 0, add, 1, sub, 64, h, l, l, r)
      SR_BF_2R( 4, 0, add, 1, sub, 64, h, l)
      SR_BF_2R(52, 0, add, 1, sub, 64, h, l)
      "addl "     "$(8*2), "  "%2\n\t"
      /* 11,  9 */
      /* %0 = data+64   %1 = data+156 */
      SR_BF_1R( 8, 0, add, 1, sub, 64, h, l, l, r)
      SR_BF_2R( 4, 0, add, 1, sub, 64, h, l)
      SR_BF_2R(64, 0, add, 1, sub, 64, h, l)
      /* 7, 5 */
      /* %0 = data+140  %1 = data+80 */
      SR_BF_k_eq_N8_SETUP
      SR_BF_k_eq_N8_2R( 4, 0, sub, 1, add, 64)
      SR_BF_k_eq_N8_2R( 8, 0, sub, 1, add, 64)
      SR_BF_k_eq_N8_1R(76, 0, add, 1, sub, 64, r, l, 48, 47)
      /* 3, 1 */
      /* %0 = data+204  %1 = data+16 */
      SR_BF_k_eq_0_2I( 4, 1, add, 0, sub, 64)
      SR_BF_k_eq_0_2I( 8, 1, add, 0, sub, 64)
      SR_BF_k_eq_0_1I(28, 1, sub, 0, add, 64, r, l)
      /* %0 = data+220  %1 = data */
      
      "addl "     "$(16*2), " "%0\n\t"
      /* %0 = data+236  %1 = data */
      "STAGE4_LOOP1:\n\t"
      STAGE_OUTPUT_1R2I( 4, 1, add, 0, sub, 32, l, r, 48)
      STAGE_OUTPUT___2I( 4, 1, add, 0, sub, 32)
      STAGE_OUTPUT_2R__( 4, 1, add, 0, sub, 32)
      STAGE_OUTPUT_2R1I(52, 1, add, 0, sub, 32, l, r)
      "dec "                  "%3\n\t"
      "jnz "      "STAGE4_LOOP1\n\t"
      /* %0 = data-20   %1 = data+256 */

   /* STAGE 5: 32pt-DFTs
      registers used: %0 = data
                      %1 = data+236
                      %2 = IDFT_CONST+16
                      %3 = 6
                      %4 = 0x5A815A81 */
   /*  op_code     src         dest */
      "addl "     "$(20*2), " "%0\n\t"
      "subl "     "$(20*2), " "%1\n\t"
      "addl "     "$(8*2), "  "%2\n\t"
      "addl "     "$6, "      "%3\n\t"
      /* 31, 29
          :   :
         11,  9 */
      /* %0 = data      %1 = data+236 */
      "STAGE5_LOOP1:\n\t"
      SR_BF_1R( 4, 0, add, 1, sub, 32, l, h, r, l)
      SR_BF_2R(28, 0, add, 1, sub, 32, l, h)
      "addl "     "$(8*2), "  "%2\n\t"
      "dec "                  "%3\n\t"
      "jnz "      "STAGE5_LOOP1\n\t"
      /* %0 = data+192  %1 = data+44 */
      /* 7, 5 */
      "addl "     "$(4*2), "  "%0\n\t"
      "subl "     "$(4*2), "  "%1\n\t"
      /* %0 = data+196  %1 = data+40 */
      SR_BF_k_eq_N8_SETUP
      SR_BF_k_eq_N8_2R( 4, 0, sub, 1, add, 32)
      SR_BF_k_eq_N8_1R(36, 0, add, 1, sub, 32, l, r, 16, 17)
      /* 3, 1 */
      /* %0 = data+228  %1 = data+8 */
      SR_BF_k_eq_0_2I( 4, 1, add, 0, sub, 32)
      SR_BF_k_eq_0_1I(12, 1, sub, 0, add, 32, l, r)
      /* %0 = data+236  %1 = data+0 */

      "addl "     "$(8*2), "  "%0\n\t"
      "addl "     "$8, "      "%3\n\t"
      /* %0 = data+244  %1 = data */
      "STAGE5_LOOP2:\n\t"
      STAGE_OUTPUT_1R2I( 4, 1, add, 0, sub, 16, r, l, 16)
      STAGE_OUTPUT_2R1I(28, 1, add, 0, sub, 16, r, l)
      "dec "                  "%3\n\t"
      "jnz "      "STAGE5_LOOP2\n\t"
      /* %0 = data-12   %1 = data+256 */

   /* STAGE 6: 64pt-DFTs
      registers used: %0 = data+4
                      %1 = data+240
                      %2 = IDFT_CONST+64
                      %3 = 14
                      %4 = 0x5A815A81 */
   /*  op_code     src         dest */
      "addl "     "$(16*2), " "%0\n\t"
      "subl "     "$(16*2), " "%1\n\t"
      "addl "     "$14, "     "%3\n\t"
      /* 1, 3 */
      /* %0 = data+4    %1 = data+240 */
      SR_BF_k_eq_0_1I(16, 0, add, 1, sub, 16, r, l)
      /* 5, 7 */
      /* %0 = data+20   %1 = data+224 */
      SR_BF_k_eq_N8_SETUP
      SR_BF_k_eq_N8_1R(16, 1, sub, 0, add, 16, r, l, 48, 47)
      /*  9, 11
          :   :
         61, 63  */
      /* %0 = data+36   %1 = data+208 */
      "STAGE6_LOOP1:\n\t"
      SR_BF_1R(16, 1, sub, 0, add, 16, h, l, l, r)
      "addl "     "$(8*2), "  "%2\n\t"
      "dec "                  "%3\n\t"
      "jnz "      "STAGE6_LOOP1\n\t"
      /* %0 = data+260  %1 = data-16 */

      "addl "     "$(16*2), " "%1\n\t"
      "movq "     "%%mm7, "   "%%mm5\n\t"
      "pcmpgtd "  "%%mm4, "   "%%mm4\n\t"
      "psrlq "    "$32, "     "%%mm5\n\t"
      "addl "     "$16, "     "%3\n\t"
      "pcmpeqw "  "%%mm5, "   "%%mm4\n\t"
      /* %0 = data+260  %1 = data */
      "STAGE6_LOOP2:\n\t"
      STAGE6_OUTPUT_01(16, 1, add, (%1), 8(%1))
      "dec "                  "%3\n\t"
      "jnz "      "STAGE6_LOOP2\n\t"
      /* %0 = data+260  %1 = data+256 */
      "psrlq "    "$16, "     "%%mm4\n\t"
      "subl "     "$(12*2), " "%0\n\t"
      "addl "     "$16, "     "%3\n\t"
      "psllq "    "$48, "     "%%mm5\n\t"
      /* %0 = data+248  %1 = data+256 */
      "STAGE6_LOOP3:\n\t"
      STAGE6_OUTPUT_01(16, 0, sub, 8(%0), (%0))
      "dec "                  "%3\n\t"
      "jnz "      "STAGE6_LOOP3\n\t"
      /* %0 = data-8    %1 = data+256 */
      "movq "     "%%mm5, "   "%%mm4\n\t"
      "addl "     "$(16*2), " "%0\n\t"
      "subl "     "$(16*2), " "%1\n\t"
      "psrlq "    "$48, "     "%%mm4\n\t"
      "addl "     "$16, "     "%3\n\t"
      /* %0 = data+8    %1 = data+240 */
      "STAGE6_LOOP4:\n\t"
      STAGE6_OUTPUT_23(16, 0, add, 1, sub)
      "dec "                  "%3\n\t"
      "jnz "      "STAGE6_LOOP4\n\t"
      /* %0 = data+264  %1 = data-16 */

   /* STAGE 7: 128pt-DFTs
      registers used: %0 = data+0
                      %1 = data+248
                      %2 = IDFT_CONST+176
                      %3 = 30
                      %4 = 0x5A815A81 */
   /*  op_code     src         dest */
      "subl "     "$(264*2), ""%0\n\t"
      "addl "     "$(264*2), ""%1\n\t"
      "addl "     "$30, "     "%3\n\t"
      /* 1, 3 */
      /* %0 = data+0    %1 = data+248 */
      SR_BF_k_eq_0_1I(8, 0, add, 1, sub, 8, l, r)
      /* 5, 7 */
      /* %0 = data+8    %1 = data+240 */
      SR_BF_k_eq_N8_SETUP
      SR_BF_k_eq_N8_1R(8, 1, sub, 0, add, 8, l, r, 16, 17)
      /*  9, 11
          :   :
        125,127  */
      /* %0 = data+16   %1 = data+232 */
      "STAGE7_LOOP1:\n\t"
      SR_BF_1R(8, 1, sub, 0, add, 8, l, h, r, l)
      "addl "     "$(8*2), "  "%2\n\t"
      "dec "                  "%3\n\t"
      "jnz "      "STAGE7_LOOP1\n\t"
      /* %0 = data+256  %1 = data-8 */

      "addl "     "$(8*2), "  "%1\n\t"
      "subl "     "$(4*2), "  "%0\n\t"
      "addl "     "$16, "     "%3\n\t"
      /* %0 = data+252  %1 = data */
      "STAGE7_LOOP2:\n\t"
      STAGE7_OUTPUT(4, 1, add, 0, sub)
      STAGE7_OUTPUT(4, 0, sub, 1, add)
      "dec "                  "%3\n\t"
      "jnz "      "STAGE7_LOOP2\n\t"
      /* %0 = data+124  %1 = data+128 */

   /* Undo Bit-Reverse Order
      registers used: %0 = data+0
                      %1 = data+92
                      %2 = IDFT_CONST+416
                      %3 = 4 */
   /*  op_code     src         dest */
      "subl "     "$(124*2), ""%0\n\t"
      "subl "     "$(36*2), " "%1\n\t"
      "addl "     "$4, "      "%3\n\t"
      /* %0 = data+0    %1 = data+92 */
      "SWAP_LOOP1:\n\t"
      SWAP8_1(0)
      "dec "                  "%3\n\t"
      "jnz "      "SWAP_LOOP1\n\t"
      /* %0 = data+116  %1 = data+92 */
      "addl "     "$12, "     "%3\n\t"
      "SWAP_LOOP2:\n\t"
      SWAP8_2(1, 0)
      "dec "                  "%3\n\t"
      "jnz "      "SWAP_LOOP2\n\t"

   /* Post-IDFT Scaling:
      registers used: %0 = data
                      %1 = data+252
                      %3 = 32 */
   /*  op_code     src         dest */
      "pcmpeqd "  "%%mm7, "   "%%mm7\n\t"
      "addl "     "$32, "     "%3\n\t"
      "psrlw "    "$15, "     "%%mm7\n\t"    /* mm7 = 0x0001000100010001 */
      /* %0 = data      %1 = data+252 */
      "POST_LOOP:\n\t"
      PRE_POST_SCALING(4, 2, 1)
      "dec "                  "%3\n\t"
      "jnz "      "POST_LOOP\n\t"
      
      "emms\n\t"
   /* output list */
   :
   /* input list */
   : "q" (data),
     "q" (data+252),
     "q" (MMX_IDFT_CONST),
     "r" (32),
     "r" (0x5A815A81)
   );
}

#undef STAGE_OUTPUT_1R2I
#undef STAGE_OUTPUT___2I
#undef STAGE_OUTPUT_2R2I
#undef STAGE_OUTPUT_2R__
#undef STAGE6_OUTPUT_01
#undef STAGE6_OUTPUT_23
#undef STAGE7_OUTPUT
#undef SR_BF_k_eq_0_1I
#undef SR_BF_k_eq_N8_SETUP
#undef SR_BF_k_eq_N8_1R
#undef SR_BF_k_eq_N8_2R
#undef SR_BF_1R
#undef SR_BF_2R
#undef SWAP8_1
#undef SWAP8_2
#undef PRE_POST_SCALING

