/***************************************************************************
                       Win32_CondVar.c  -  description
                             -------------------
   Implementation of the CONDVAR API, using the Win32 thread library.
                             -------------------
    begin                : Sat Jul 17 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdlib.h>

/* Public Functions */
/* Creates a new conditional variable. */
CONDVAR new_CONDVAR()
{
   WIN32_cond_t *wc = (WIN32_cond_t *)malloc(sizeof(WIN32_cond_t));

   if (!wc)
      return NULL;

   wc->event[0] = CreateEvent(NULL, FALSE, FALSE, NULL); // signal event
   if (!wc->event[0]) {
      free(wc);
      return NULL;
   }
   wc->event[1] = CreateEvent(NULL, TRUE,  FALSE, NULL); // broadcast event
   if (!wc->event[1]) {
      CloseHandle(wc->event[0]);
      free(wc);
      return NULL;
   }
   return (CONDVAR)wc;
}

/* Waits for the conditional variable to change.
   Returns non-zero value for success. */
int wait_CONDVAR(CONDVAR cv, MUTEX m)
{
   WIN32_cond_t *wc = (WIN32_cond_t *)cv;

   // unlock mutex, so another thread can signal event
   if (ReleaseMutex((HANDLE)m) == FALSE)
      return 0;

   // wait for event signal
   if (WaitForMultipleObjects(2, wc->event, FALSE, 5000) == WAIT_FAILED)
      return 0;

   // re-aquire lock for testing purposes
   if (WaitForSingleObject((HANDLE)m, INFINITE) != WAIT_OBJECT_0)
      return 0;

   return 1;
}

