/***************************************************************************
                          PSI_SECTION.c  -  description
                             -------------------
   Stuff needed for handling PSI SECTIONs.
                             -------------------
    begin                : Fri Aug 6 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "PSI_SECTION.h"


/* Reads the first PSI SECTION, in data_stream, into psi.
      INPUT: data_stream - Points to the first byte of the PSI SECTION.
                             This function assumes there are at most
                             8 bytes.
      OUTPUT: psi - Points to the PSI_SECTION that will hold data taken from
                      data_stream.
      RETURN: Success - A non-zero value
              Error   - 0 */
int PSI_SECTION_read(const u8 *data_stream, PSI_SECTION *psi)
{
   psi->table_id = *data_stream;
   data_stream++;
   if (psi->table_id == PSI_SECTION_FORBIDDEN_ID)
      return 0;   /* bad table_id value */

   psi->flags          = *data_stream;
   psi->payload_len16  = (u16)(*data_stream & 0x0F) << 8;
   data_stream++;

   psi->payload_len16 |= (u16)(*data_stream);
   data_stream++;
   if (psi->payload_len16 > (PSI_SECTION_MAX_SIZE-3))
      return 0;   /* section is too big */

   if (PSI_SECTION_SYNTAX_FLAG(psi)) {
      if (psi->payload_len16 < 9)
         return 0;   /* section is too small */
      psi->payload_len16 -= 9;
      
      psi->table_id_extension  = (u16)(*data_stream) << 8;
      data_stream++;

      psi->table_id_extension |= (u16)(*data_stream);
      data_stream++;

      psi->flags |= *data_stream & 0x3F;
      data_stream++;

      psi->section_number = *data_stream;
      data_stream++;

      psi->last_section_number = *data_stream;
      data_stream++;
   }

   psi->payload = data_stream;
   return 1;
}

/* Checks the CRC value of the given psi.
      INPUT: psi - Points to the PSI_SECTION that needs to be checked.
      RETURN: CRC good - A non-zero value
              CRC bad  - 0 */
int PSI_SECTION_chk_crc(PSI_SECTION *psi)
{
#ifdef __COMPILE_CRC_CODE__
   if (PSI_SECTION_SYNTAX_FLAG(psi)) {
      u32       CRC_32;
      const u8 *psi_ptr = psi->payload - 8;
               *crc_ptr = psi->payload + psi->payload_len16;

      /* Calculate CRC value of the (psi->payload_len16+8) bytes,
           starting with the byte at psi_ptr. Compare that CRC value
           with the CRC_32 value. */
           
      CRC_32  = (u32)(*crc_ptr) << 24;
      crc_ptr++;
      CRC_32 |= (u32)(*crc_ptr) << 16;
      crc_ptr++;
      CRC_32 |= (u32)(*crc_ptr) <<  8;
      crc_ptr++;
      CRC_32 |= (u32)(*crc_ptr);
   }
#endif
   return 1;
}
