/***************************************************************************
                         dvd_input.c  -  description
                             -------------------
   API to get the stream from the DVD device (using libdvdnav), instead of
    regular files.  This API mimics the one found in common/File.h
                             -------------------
    begin                : Fri Feb 3 2006
    copyright            : (C) 2006 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifdef _DVD_DISPLAY_ERRORS_
# include <stdio.h>
#endif
#include  "dvd_input.h"
#include  "MPEG2_SYS.h"
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#ifndef DVD_DEFAULT_DEVICE
# define DVD_DEFAULT_DEVICE   "/dev/dvd"
#endif
#ifndef DVD_DEFAULT_LANG
# define DVD_DEFAULT_LANG     "en"
#endif

#ifdef _DVD_DISPLAY_ERRORS_
const char   *dvd_err_str        = "DVDNAV ERROR: %s\n";
#endif
const char   *dvd_device_default = DVD_DEFAULT_DEVICE;
char          dvd_lang[4]        = DVD_DEFAULT_LANG;
u16           dvd_title_last     = 0xFFFF,
              dvd_title          = 1;
DVD_BLK_COUNT dvd_chapter        = 1,
              dvd_blk            = 0,
              dvd_skip_blks      = 0;
CLK_VAL       dvd_time           = 0;


DVD_PTR dvd_input_open(const char *device)
{
   DVD_PTR ptr;
   int32_t i, j,
           titles = 0;
   
   if (dvdnav_open(&ptr, device) == DVDNAV_STATUS_OK) {
      if ( (dvdnav_set_readahead_flag(ptr, 0)           == DVDNAV_STATUS_OK)&&
           (dvdnav_set_PGC_positioning_flag(ptr, 1)     == DVDNAV_STATUS_OK)&&
           (dvdnav_get_next_block(ptr, sys.buf, &i, &j) == DVDNAV_STATUS_OK)&&
           (dvdnav_get_number_of_titles(ptr, &titles)   == DVDNAV_STATUS_OK)&&
           (titles > 0) ) {
/*
   dvdnav_menu_language_select(ptr, dvd_lang);
   dvdnav_audio_language_select(ptr, dvd_lang);
   dvdnav_spu_language_select(ptr, dvd_lang);
*/
         if ((int32_t)dvd_title_last > titles)
            dvd_title_last = (u16)titles;
         if ((int32_t)dvd_title > titles)
            dvd_title      = (u16)titles;
         DVD_REWIND(ptr);
         return ptr;
      }
#ifdef _DVD_DISPLAY_ERRORS_
      fprintf(stderr, dvd_err_str, dvdnav_err_to_string(ptr));
#endif
      DVD_CLOSE(ptr);
  }
  return DVD_PTR_BAD_VALUE;
}

DVD_BLK_COUNT dvd_input_read(DVD_PTR dvd, u8 *buf, DVD_BLK_COUNT buf_size)
{
   int32_t       event, len;
   DVD_BLK_COUNT blks_read = 0;
   
   do {
      if (dvdnav_get_next_block(dvd, buf, &event, &len) == DVDNAV_STATUS_ERR) {
#ifdef _DVD_DISPLAY_ERRORS_
         fprintf(stderr, dvd_err_str, dvdnav_err_to_string(dvd));
#endif
         break;
      }
      switch (event) {
      case DVDNAV_BLOCK_OK:
         dvd_blk++;
         if (!dvd_skip_blks) {
            blks_read++;
            buf += DVD_VIDEO_LB_LEN;
         }
         else
            dvd_skip_blks--;
         break;
      case DVDNAV_STILL_FRAME:
         dvdnav_still_skip(dvd);
         break;
      case DVDNAV_WAIT:
         dvdnav_wait_skip(dvd);
         break;
      case DVDNAV_NAV_PACKET:
      {
         pci_t *pci = dvdnav_get_current_nav_pci(dvd);
         if (pci) {
#if (MPEG2_SYS_CLK_FREQ >= 90000)
            dvd_time = (CLK_VAL)pci->pci_gi.vobu_s_ptm * (MPEG2_SYS_CLK_FREQ/90000);
#else
            dvd_time = (CLK_VAL)pci->pci_gi.vobu_s_ptm / (90000/MPEG2_SYS_CLK_FREQ);
#endif
/*
            dvd_time  = (CLK_VAL)pci->pci_gi.e_eltm.hour * (3600*MPEG2_SYS_CLK_FREQ);
            dvd_time += (CLK_VAL)pci->pci_gi.e_eltm.minute * (60*MPEG2_SYS_CLK_FREQ);
            dvd_time += (CLK_VAL)pci->pci_gi.e_eltm.second *     MPEG2_SYS_CLK_FREQ;
*/
         }
         break;
      }
      case DVDNAV_VTS_CHANGE:
      case DVDNAV_CELL_CHANGE:
         if (dvdnav_current_title_info(dvd, &event, &len) == DVDNAV_STATUS_OK) {
            if ((int32_t)dvd_title == event) {
               dvd_chapter = (DVD_BLK_COUNT)len;
               dvd_blk     = 0;
               break;
            }
         }
#ifdef _DVD_DISPLAY_ERRORS_
         else
            fprintf(stderr, dvd_err_str, dvdnav_err_to_string(dvd));
#endif
      case DVDNAV_STOP:
         buf_size = 0;
      }
   } while (blks_read < buf_size);
   return blks_read;
}

DVD_BLK_COUNT DVD_NUM_CHAPTERS(DVD_PTR dvd)
{
   DVD_BLK_COUNT num = 0;

   dvdnav_get_number_of_parts(dvd, (int32_t)dvd_title, (int32_t *)&num);
   return num;
}

u16 DVD_AUDIO_LANG(DVD_PTR dvd, u16 PID)
{
   if ((PID&0x00FF) == 0x00BD)
      PID >>= 8;
   return dvdnav_audio_stream_to_lang(dvd, (uint8_t)(PID & 0x001F));
}

u16 DVD_AUDIO_DEFAULT(DVD_PTR dvd)
{
   u16  lang, 
        PID = 0;
   
   while ((lang = dvdnav_audio_stream_to_lang(dvd, (uint8_t)PID)) != 0xFFFF) {
      if ( (dvd_lang[0] == (char)(lang>>8))&&
           (dvd_lang[1] == (char)(lang&0x00FF)) )
         return PID;
      PID++;
   }
#ifdef _DVD_DISPLAY_ERRORS_
   fprintf(stderr, "DVDNAV WARN: Cannot find language \"%s\".\n", dvd_lang);
#endif
   return 0;
}

