/***************************************************************************
                  mmx_msvc_motion_vector.c  -  description
                             -------------------
   MMX versions of the motion vector functions (MSVC inline assembly).
                             -------------------
    begin                : Tue Dec 28 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/* Copy a frame. */
void MV_COPY_FRAME(void *to, void *from, u32 _size)
{  __asm {
      mov   eax,  dword ptr[to]
      mov   ebx,  dword ptr[from]
      mov   ecx,  _size
   MVCF_LOOP:
      movq  mm0,     [ebx]
      movq  [eax],   mm0
      add   ebx,     8
      add   eax,     8
      sub   ecx,     8
      jg    MVCF_LOOP
      emms
}  }

/* Clear a block. */
void MV_CLEAR_BLOCK(s16 *blk)
{  __asm {
      mov   eax,  dword ptr[blk]
      mov   ecx,  8
      pxor  mm7,  mm7
   MVCB_LOOP:
      movq  [eax],   mm7
      movq  [eax+8], mm7
      add   eax,     16
      dec   ecx
      jnz   MVCB_LOOP
}  }


/* The MV_PRED_?? functions are used to get the predictors from
     the reference block(s). */
void MV_PRED___(s16 *inp, u8 *ref, u32 _inc)
{  __asm {
      mov   eax,  dword ptr[inp]
      mov   ebx,  dword ptr[ref]
      mov   ecx,  8
      mov   edx,  _inc
      pxor  mm7,  mm7
   MVP_LOOP:
      movq        mm2,     [ebx]
      movq        mm0,     [eax]
      movq        mm3,     mm2
      punpcklbw   mm2,     mm7
      movq        mm1,     [eax+8]
      punpckhbw   mm3,     mm7
      paddsw      mm0,     mm2
      paddsw      mm1,     mm3
      movq        [eax],   mm0
      add         ebx,     edx
      movq        [eax+8], mm1
      add         eax,     16
      dec         ecx
      jnz         MVP_LOOP
}  }

void MV_PRED_X_(s16 *inp, u8 *ref, u32 _inc)
{  __asm {
      mov   eax,  dword ptr[inp]
      mov   ebx,  dword ptr[ref]
      mov   ecx,  8
      mov   edx,  _inc
      pxor  mm7,  mm7
   MVP_X_LOOP:
      movq        mm2,     [ebx]
      movq        mm0,     [eax]
      movq        mm3,     mm2
      punpcklbw   mm2,     mm7
      movq        mm1,     [eax+8]
      punpckhbw   mm3,     mm7
      movq        mm4,     [ebx+1]
      movq        mm5,     mm4
      punpcklbw   mm4,     mm7
      punpckhbw   mm5,     mm7
      paddsw      mm2,     mm4
      paddsw      mm3,     mm5
      psraw       mm2,     1
      psraw       mm3,     1
      paddsw      mm0,     mm2
      paddsw      mm1,     mm3
      movq        [eax],   mm0
      add         ebx,     edx
      movq        [eax+8], mm1
      add         eax,     16
      dec         ecx
      jnz         MVP_X_LOOP
}  }

void MV_PRED__Y(s16 *inp, u8 *ref, u32 _inc, u32 inc_y)
{  __asm {
      mov   edi,  dword ptr[inp]
      mov   esi,  dword ptr[ref]
      mov   eax,  inc_y
      mov   ebx,  _inc
      mov   ecx,  8
      add   eax,  esi
      pxor  mm7,  mm7
   MVP_Y_LOOP:
      movq        mm2,     [esi]
      movq        mm0,     [edi]
      movq        mm3,     mm2
      punpcklbw   mm2,     mm7
      movq        mm1,     [edi+8]
      punpckhbw   mm3,     mm7
      movq        mm4,     [eax]
      movq        mm5,     mm4
      punpcklbw   mm4,     mm7
      punpckhbw   mm5,     mm7
      paddsw      mm2,     mm4
      paddsw      mm3,     mm5
      psraw       mm2,     1
      psraw       mm3,     1
      paddsw      mm0,     mm2
      paddsw      mm1,     mm3
      add         esi,     ebx
      movq        [edi],   mm0
      add         eax,     ebx
      movq        [edi+8], mm1
      add         edi,     16
      dec         ecx
      jnz         MVP_Y_LOOP
}  }

void MV_PRED_XY(s16 *inp, u8 *ref, u32 _inc, u32 inc_y)
{  __asm {
      mov   edi,  dword ptr[inp]
      mov   esi,  dword ptr[ref]
      mov   eax,  inc_y
      mov   ebx,  _inc
      mov   ecx,  8
      add   eax,  esi
      pxor  mm7,  mm7
   MVP_XY_LOOP:
      movq        mm2,     [esi]
      movq        mm0,     [edi]
      movq        mm3,     mm2
      punpcklbw   mm2,     mm7
      movq        mm1,     [edi+8]
      punpckhbw   mm3,     mm7
      movq        mm4,     [esi+1]
      movq        mm5,     mm4
      punpcklbw   mm4,     mm7
      punpckhbw   mm5,     mm7
      paddsw      mm2,     mm4
      paddsw      mm3,     mm5
      movq        mm4,     [eax]
      movq        mm5,     mm4
      punpcklbw   mm4,     mm7
      punpckhbw   mm5,     mm7
      paddsw      mm2,     mm4
      paddsw      mm3,     mm5
      movq        mm4,     [eax+1]
      movq        mm5,     mm4
      punpcklbw   mm4,     mm7
      punpckhbw   mm5,     mm7
      paddsw      mm2,     mm4
      paddsw      mm3,     mm5
      psraw       mm2,     2
      psraw       mm3,     2
      paddsw      mm0,     mm2
      paddsw      mm1,     mm3
      add         esi,     ebx
      movq        [edi],   mm0
      add         eax,     ebx
      movq        [edi+8], mm1
      add         edi,     16
      dec         ecx
      jnz MVP_XY_LOOP
}  }


/* The MV_FIELD_PRED_?? functions are similar to the MV_PRED_?? functions,
     except these functions are used when different reference blocks are
     needed for each field. */
void MV_FIELD_PRED___(s16 *inp0, u32 inp_inc, 
                      u8  *ref0, u32 _inc)
{  __asm {
      mov   edi,  dword ptr[inp0]
      mov   esi,  dword ptr[ref0]
      mov   eax,  _inc
      mov   ebx,  inp_inc
      mov   ecx,  4
      shl   ebx,  1
      pxor  mm7,  mm7
   MVFP_LOOP:
      movq        mm2,     [esi]
      movq        mm0,     [edi]
      movq        mm3,     mm2
      punpcklbw   mm2,     mm7
      movq        mm1,     [edi+8]
      punpckhbw   mm3,     mm7
      paddsw      mm0,     mm2
      paddsw      mm1,     mm3
      movq        [edi],   mm0
      add         esi,     eax
      movq        [edi+8], mm1
      add         edi,     ebx
      dec         ecx
      jnz         MVFP_LOOP
}  }
   
void MV_FIELD_PRED_X_(s16 *inp0, u32 inp_inc, 
                      u8  *ref0, u32 _inc)
{  __asm {
      mov   edi,  dword ptr[inp0]
      mov   esi,  dword ptr[ref0]
      mov   eax,  _inc
      mov   ebx,  inp_inc
      mov   ecx,  4
      shl   ebx,  1
      pxor  mm7,  mm7
   MVFP_X_LOOP:
      movq        mm2,     [esi]
      movq        mm0,     [edi]
      movq        mm3,     mm2
      punpcklbw   mm2,     mm7
      movq        mm1,     [edi+8]
      punpckhbw   mm3,     mm7
      movq        mm4,     [esi+1]
      movq        mm5,     mm4
      punpcklbw   mm4,     mm7
      punpckhbw   mm5,     mm7
      paddsw      mm2,     mm4
      paddsw      mm3,     mm5
      psraw       mm2,     1
      psraw       mm3,     1
      paddsw      mm0,     mm2
      paddsw      mm1,     mm3
      movq        [edi],   mm0
      add         esi,     eax
      movq        [edi+8], mm1
      add         edi,     ebx
      dec         ecx
      jnz         MVFP_X_LOOP
}  }

void MV_FIELD_PRED__Y(s16 *inp0, u32 inp_inc, 
                      u8  *ref0, u32 _inc, u32 inc_y)
{  __asm {
      mov   edi,  dword ptr[inp0]
      mov   esi,  dword ptr[ref0]
      mov   edx,  inc_y
      mov   eax,  _inc
      mov   ebx,  inp_inc
      mov   ecx,  4
      shl   ebx,  1
      add   edx,  esi
      pxor  mm7,  mm7
   MVFP_Y_LOOP:
      movq        mm2,     [esi]
      movq        mm0,     [edi]
      movq        mm3,     mm2
      punpcklbw   mm2,     mm7
      movq        mm1,     [edi+8]
      punpckhbw   mm3,     mm7
      movq        mm4,     [edx]
      movq        mm5,     mm4
      punpcklbw   mm4,     mm7
      punpckhbw   mm5,     mm7
      paddsw      mm2,     mm4
      paddsw      mm3,     mm5
      psraw       mm2,     1
      psraw       mm3,     1
      paddsw      mm0,     mm2
      paddsw      mm1,     mm3
      add         esi,     eax
      movq        [edi],   mm0
      add         edx,     eax
      movq        [edi+8], mm1
      add         edi,     ebx
      dec         ecx
      jnz         MVFP_Y_LOOP
}  }

void MV_FIELD_PRED_XY(s16 *inp0, u32 inp_inc, 
                      u8  *ref0, u32 _inc, u32 inc_y)
{  __asm {
      mov   edi,  dword ptr[inp0]
      mov   esi,  dword ptr[ref0]
      mov   edx,  inc_y
      mov   eax,  _inc
      mov   ebx,  inp_inc
      mov   ecx,  4
      shl   ebx,  1
      add   edx,  esi
      pxor  mm7,  mm7
   MVFP_XY_LOOP:
      movq        mm2,     [esi]
      movq        mm0,     [edi]
      movq        mm3,     mm2
      punpcklbw   mm2,     mm7
      movq        mm1,     [edi+8]
      punpckhbw   mm3,     mm7
      movq        mm4,     [esi+1]
      movq        mm5,     mm4
      punpcklbw   mm4,     mm7
      punpckhbw   mm5,     mm7
      paddsw      mm2,     mm4
      paddsw      mm3,     mm5
      movq        mm4,     [edx]
      movq        mm5,     mm4
      punpcklbw   mm4,     mm7
      punpckhbw   mm5,     mm7
      paddsw      mm2,     mm4
      paddsw      mm3,     mm5
      movq        mm4,     [edx+1]
      movq        mm5,     mm4
      punpcklbw   mm4,     mm7
      punpckhbw   mm5,     mm7
      paddsw      mm2,     mm4
      paddsw      mm3,     mm5
      psraw       mm2,     2
      psraw       mm3,     2
      paddsw      mm0,     mm2
      paddsw      mm1,     mm3
      add         esi,     eax
      movq        [edi],   mm0
      add         edx,     eax
      movq        [edi+8], mm1
      add         edi,     ebx
      dec         ecx
      jnz         MVFP_XY_LOOP
}  }


/* Saturates the final pixel values, in the block, and stores them in
     the frame buffer. */
void MV_SATURATION(s16 *blk, u8 *outp, int _inc)
{  __asm {
      mov         esi,  dword ptr[blk]
      mov         edi,  dword ptr[outp]
      mov         ebx,  _inc
      mov         ecx,  0x80808080
      movd        mm6,  ecx
      pxor        mm7,  mm7
      mov         ecx,  8
      punpcklbw   mm6,  mm7
   MV_SATURATION_LOOP:
      movq        mm0,     [esi]
      movq        mm1,     [esi+8]
      paddsw      mm0,     mm6
      paddsw      mm1,     mm6
      movq        mm2,     mm0
      movq        mm3,     mm1
      pcmpgtw     mm2,     mm7
      pcmpgtw     mm3,     mm7
      pand        mm0,     mm2
      pand        mm1,     mm3
      packuswb    mm0,     mm1
      add         esi,     16
      movq        [edi],   mm0
      add         edi,     ebx
      dec         ecx
      jnz         MV_SATURATION_LOOP
      emms
}  }


/* Adds the predictor block to the DCT block, to obtain the final pixel
     values.  The final values are stored in the frame buffer. */
void MV_ADD_PRED(s16 *blk, s16 *p_blk, u8 *outp, 
                 int  _inc, u32  sh)
{  __asm {
      mov   esi,  dword ptr[blk]
      mov   eax,  dword ptr[p_blk]
      mov   edi,  dword ptr[outp]
      mov   ebx,  _inc
      mov   ecx,  sh
      movd  mm4,  ecx
      pxor  mm7,  mm7
      mov   ecx,  8
   MV_ADD_PRED_LOOP:
      movq        mm0,     [eax]
      movq        mm1,     [eax+8]
      psraw       mm0,     mm4
      psraw       mm1,     mm4
      paddsw      mm0,     [esi]
      paddsw      mm1,     [esi+8]
      movq        mm2,     mm0
      movq        mm3,     mm1
      pcmpgtw     mm2,     mm7
      pcmpgtw     mm3,     mm7
      pand        mm0,     mm2
      pand        mm1,     mm3
      add         esi,     16
      packuswb    mm0,     mm1
      add         eax,     16
      movq        [edi],   mm0
      add         edi,     ebx
      dec         ecx
      jnz         MV_ADD_PRED_LOOP
      emms
}  }


/* Copies the predictor block into the frame buffer. */
void MV_COPY_PRED(s16 *p_blk, u8 *outp, int _inc, u32 sh)
{  __asm {
      mov   esi,  dword ptr[p_blk]
      mov   edi,  dword ptr[outp]
      mov   ebx,  _inc
      mov   ecx,  sh
      movd  mm4,  ecx
      pxor  mm7,  mm7
      mov   ecx,  8
   MV_COPY_PRED_LOOP:
      movq        mm0,     [esi]
      movq        mm1,     [esi+8]
      psraw       mm0,     mm4
      psraw       mm1,     mm4
      movq        mm2,     mm0
      movq        mm3,     mm1
      pcmpgtw     mm2,     mm7
      pcmpgtw     mm3,     mm7
      pand        mm0,     mm2
      pand        mm1,     mm3
      packuswb    mm0,     mm1
      add         esi,     16
      movq        [edi],   mm0
      add         edi,     ebx
      dec         ecx
      jnz         MV_COPY_PRED_LOOP
      emms
}  }

