/***************************************************************************
                            wmv.h  -  description
                             -------------------
   Stuff needed for handling wmv demuxing.
                             -------------------
    begin                : Thu May 25 2006
    copyright            : (C) 2006 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __WMV_H__
#define __WMV_H__

#include "common/typedefs.h"
#include "SEQ_LAYER.h"


#define WMV_STATE_MASK        0xFF000000
#define WMV_next_guid_STATE   0x00000000
#define WMV_GUID_HEADER       0x01000000
#define WMV_GUID_FILE         0x02000000
#define WMV_GUID_STREAM       0x03000000
#define WMV_GUID_DATA         0x04000000
#define WMV_FRAME_FRAG_STATE  0x05000000
#define WMV_skip_guid_STATE   0x06000000

#define WMV_END_FRAME_STATE   (WMV_GUID_DATA+40)

#define WMV_BROADCAST   1
#define WMV_LOCAL       2


#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
   const
   u8 *payload;
   u32 payload_len,        /* Number of bytes in the payload */   
       state, guid;
   u64 size,
       FH_pkts;
   u32 FH_flags, FH_min_pkt, FH_max_pkt,
       SH_size,
       WD_sbytes, WD_pkt_size, WD_pad_size,
       WD_frags, WD_frag_size, WD_frag_offset,
       WD_frame_size, WD_hdr_size,
       frame_size, frag_offset, skip_frames;
   u8  stream_id, cur_id, pad1, pad2;
} WMV_DATA;

/* Read the next GUID.
      INPUT: wmv - Pointer to a WMV_DATA structure. */
void WMV_next_guid(WMV_DATA *wmv);

/* Read the FILE_HEADER.
      INPUT: wmv - Pointer to a WMV_DATA structure. */
void WMV_read_file(WMV_DATA *wmv);

/* Read the STREAM_HEADER.
      INPUT: wmv - Pointer to a WMV_DATA structure.
             seq - Pointer to a SEQ_LAYER structure. */
void WMV_read_stream(WMV_DATA *wmv, SEQ_LAYER *seq);

/* Read the WMV_DATA.
      INPUT: wmv - Pointer to a WMV_DATA structure. */
void WMV_read_data(WMV_DATA *wmv);

/* Skip the current GUID.
      INPUT: wmv - Pointer to a WMV_DATA structure. */
void WMV_skip_guid(WMV_DATA *wmv);

#ifdef __cplusplus
};
#endif

#endif

